/***
*
*  Ibrowse.prg
*
*  Illustration of IMAGE BROWSE with Light Lib Images.
*
*  Copyright (c) 1993, DFL.
*  All rights reserved.
*
*  Compile:  CLIPPER iBrowse /m /n /w
*  Link:     RTLINK FILE iBrowse, llibitoo, llibgtoo, getsys LIB LLIBI,LLIBG,CLIPPER
*  Execute:  iBrowse <ImgFile>
*  Note :    .BMP icons or iBrowse.bml must be present in the same directory
*/

#include "Inkey.ch"

#include "llibi.ch"          //  Use Light Lib Images defines
#include "llibitoo.ch"       //  Use Light Lib Images tools defines
#include "llibg.ch"          //  Use Light Lib Graphics defines
#include "llibgtoo.ch"       //  Use Light Lib Graphics tools defines


*
FUNCTION iBrowse( cImgFileSpec )

   LOCAL nUserMode := 0      // Graphics mode choosen by user
   LOCAL nI        := 0      // Indice
   LOCAL aImg      := {}

   CLS
   SET CURSOR OFF            // Hide cursor

   IF (nUserMode := nChooseVideoMode())==0

      QUIT                // User strike ESC or no graphics modes available

   ELSE

      gMode( nUserMode )  // Turn in the choosen graphics mode
                          // for exemple gMode(LLG_VIDEO_VGA_640_480_16)

                          // Display Wall Paper with WallPaper(FileName)
                          // This little CLIPPER function
                          // use LLIBG calls. See LLibgToo.prg

      IF gMode()[LLG_MODE_COLOR_MAX]==256
                          // If 256 color are available display a
                          // 256 colors BMP
         WallPaper("ATLAN.BMP")

      ELSE                // Else display a 16 colors BMP

         WallPaper("LEAVES.BMP")

      ENDIF

   ENDIF

   IF VALTYPE(cImgFileSpec)<>'C'

                             // Assume it was a problem with the params
      ?? "Usage: iBrowse <cImgFileSpec>"
      INKEY(0)

   ELSE
                             // Load specified files
      aImg := DIRECTORY(cImgFileSpec)

                             // Extract filenames
      FOR nI := 1 TO LEN(aImg)
          aImg[nI]:=aImg[nI,1]
      NEXT
      ASORT(aImg)

      IF LEN(aImg)>0         // If images are available

                             // Browse array of images
         ImageBrowse(02, 05, MAXROW()-2, MAXCOL()-2,aImg)

      ELSE

         ?? "Error: No files available"
         INKEY(0)

      ENDIF

   ENDIF

   QUIT

   RETURN (NIL)



*
FUNCTION ImageBrowse(nTop, nLeft, nBottom, nRight, aImg)


   LOCAL nKey                // Keystroke
   LOCAL lMore := .T.        // Loop control

   LOCAL aWinButtons  := {}  //  Arrays to handle buttons
   LOCAL aVertScroll  := {}  //  and scrolls bars
   LOCAL aHoriScroll  := {}  //

                             // General savings
   LOCAL xScreen     := SAVESCREEN( nTop, nLeft, nBottom, nRight)
   LOCAL cColorSave  := SETCOLOR('N/W,W+/G,W,W,GR+/W')
   LOCAL nCursoSave  := SETCURSOR(0)

                             // Browse pseudo object
   LOCAL oBrowseImg  := ARRAY(BROWSE_IMG_SIZE)

   oBrowseImg:lIsStable      := .F.
   oBrowseImg:nCurrentX      := 0
   oBrowseImg:nCurrentY      := 0
   oBrowseImg:nScaleX        := 0
   oBrowseImg:nScaleY        := 0

                             // No image loaded, ptr are empty
   oBrowseImg:aOriPtr        := {}
   oBrowseImg:aDisPtr        := {}

                             // Compute display window pixels values
   oBrowseImg:nWinDispX1     := FONT_SIZE_X * ( nLeft + 2 )
   oBrowseImg:nWinDispY1     := FONT_SIZE_Y * ( nTop + 3 )
   oBrowseImg:nWinDispX2     := FONT_SIZE_X * ( nRight - 3 )  - 1
   oBrowseImg:nWinDispY2     := FONT_SIZE_Y * ( nBottom - 1 ) - 1
   oBrowseImg:nWinDispWidth  := oBrowseImg:nWinDispX2 - oBrowseImg:nWinDispX1 + 1
   oBrowseImg:nWinDispHeight := oBrowseImg:nWinDispY2 - oBrowseImg:nWinDispY1 + 1

                             // Fill the images to be browsed
   oBrowseImg:aImgAvail      := aImg
   oBrowseImg:nImgInUse      := 0

                             // Set the block to standard apply key
   oBrowseImg:bApplyKey      := { |oBrowse,nKey| ImgApplyKey( oBrowse, nKey ) }
   oBrowseImg:aUdfShort      := { { | nStatus , nCall | lImgGauge( nStatus , nCall, .F., oBrowseImg ) } , 10 }
   oBrowseImg:aUdfLong       := { { | nStatus , nCall | lImgGauge( nStatus , nCall, .T., oBrowseImg ) } , 20 }

                             // -LLIBG- Display 3D box
   DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Set gauge function
   iSet( LLI_SET_UDF , oBrowseImg:aUdfShort )

                             // If loading next image is a succes
   IF lLoadImgFile(oBrowseImg,+1)

                             // -LLIBG- Add vertical scroll bar
      @ (nRight-3)*FONT_SIZE_X                                                ,;
        (nTop+1)*FONT_SIZE_Y                                                  ,;
        (nRight-1)*FONT_SIZE_X                                                ,;
        (nBottom-1)*FONT_SIZE_Y                                                ;
        SCROLLBAR                                                              ;
        STYLE BUTTON_TYPE_SCROLL_VERT                                          ;
        ACTIONUP  { || NIL }                                                   ;
        ACTIONDN  { || NIL }                                                   ;
        ACTIONBAR { |nSens,nPercent| ImageVertBar(nSens,nPercent,oBrowseImg) } ;
        HANDLE aVertScroll                                                     ;
        ATTACH aWinButtons


                                // -LLIBG- Add horizontal scroll bar
      @ (nLeft+2)*FONT_SIZE_X                                                  ,;
        (nBottom-1)*FONT_SIZE_Y                                                ,;
        (nRight-3)*FONT_SIZE_X                                                 ,;
        (nBottom-0)*FONT_SIZE_Y                                                 ;
        SCROLLBAR                                                               ;
        STYLE BUTTON_TYPE_SCROLL_HORI                                           ;
        ACTIONUP  { || NIL }                                                    ;
        ACTIONDN  { || NIL }                                                    ;
        ACTIONBAR { |nSens,nPercent| ImageHoriBar(nSens,nPercent,oBrowseImg) }  ;
        HANDLE aHoriScroll                                                      ;
        ATTACH aWinButtons

                             // Add a close window button
      @ nLeft*FONT_SIZE_X                       ,;
        nTop *FONT_SIZE_Y                      ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_KEY                    ;
        ACTION K_ESC                             ;
        DISPLAYUP gBmpLoad("CLOSEWIN.BMU")       ;
        DISPLAYDN gBmpLoad("CLOSEWIN.BMD")       ;
        ATTACH aWinButtons

      IF gMode()[LLG_MODE_COLOR_MAX]==256
                             // In 256 colors mode, add a button to manage
                             // exclusive / shared palette
         @ oBrowseImg:nWinDispX2 + 1                               ,;
           oBrowseImg:nWinDispY2 + 1                               ,;
           oBrowseImg:nWinDispX2 + 16                              ,;
           oBrowseImg:nWinDispY2 + 16                               ;
           BUTTON                                                   ;
           STYLE BUTTON_TYPE_RELEASE                                ;
           ACTION { || PalExcShaSwitch(oBrowseImg) }                ;
           DISPLAYUP { || PalExcShaDisp(oBrowseImg,.T.) }           ;
           DISPLAYDN { || PalExcShaDisp(oBrowseImg,.F.) }           ;
           ATTACH aWinButtons

      ENDIF

                                // -LLIBG- Add an icon bar


                                // -LLIBG- Add a button to allow user
                                //         to quit
      @ (nLeft+2 )*FONT_SIZE_X                  ,;
        (nTop+1  )*FONT_SIZE_Y                 ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_KEY                    ;
        ACTION K_ESC                             ;
        ACCELERATOR "Q"                          ;
        DISPLAYUP gBmpLoad("QUIT.BMU")           ;
        DISPLAYDN gBmpLoad("QUIT.BMD")           ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to allow user
                                //         to load previous image
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1       ,;
        ATAIL(aWinButton)[BUTTON_TOP]          ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_RELEASE                ;
        ACTION { || lLoadImgFile(oBrowseImg,-1) };
        ACCELERATOR "P"                          ;
        DISPLAYUP gBmpLoad("GOBAK.BMU")          ;
        DISPLAYDN gBmpLoad("GOBAK.BMD")          ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to allow user
                                //         to load next image
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1       ,;
        ATAIL(aWinButton)[BUTTON_TOP]          ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_RELEASE                ;
        ACTION { || lLoadImgFile(oBrowseImg,+1) };
        ACCELERATOR "N"                          ;
        DISPLAYUP gBmpLoad("GOFOR.BMU")          ;
        DISPLAYDN gBmpLoad("GOFOR.BMD")          ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to fit image into
                                //         the window
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgZoom(ASC('/'),oBrowseImg) }                         ;
        ACCELERATOR "S"                                                    ;
        DISPLAYUP gBmpLoad("SMALEST.BMU")                                  ;
        DISPLAYDN gBmpLoad("SMALEST.BMD")                                  ;
        ATTACH aWinButtons

                                // -LLIBG- Add a button to zoom a part of
                                //         the image
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgZoom(ASC(''),oBrowseImg) }                         ;
        ACCELERATOR "A"                                                    ;
        DISPLAYUP gBmpLoad("AREA.BMU")                                     ;
        DISPLAYDN gBmpLoad("AREA.BMD")                                     ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to rotate to the right
                                //         the image
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgRotate(90,oBrowseImg) }                             ;
        ACCELERATOR "9"                                                    ;
        DISPLAYUP gBmpLoad("ROTAT090.BMU")                                 ;
        DISPLAYDN gBmpLoad("ROTAT090.BMD")                                 ;
        ATTACH aWinButtons

                                // -LLIBG- Add a button to flip the image upside
                                //         down
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgRotate(180,oBrowseImg) }                            ;
        ACCELERATOR "1"                                                    ;
        DISPLAYUP gBmpLoad("ROTAT180.BMU")                                 ;
        DISPLAYDN gBmpLoad("ROTAT180.BMD")                                 ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to allow printing
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || PrintImg(oBrowseImg) }                                 ;
        ACCELERATOR "P"                                                    ;
        DISPLAYUP gBmpLoad("PRINT.BMU")                                    ;
        DISPLAYDN gBmpLoad("PRINT.BMD")                                    ;
        ATTACH aWinButtons

                                // -LLIBG- Add a button to allow scanning
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1      ,;
        ATAIL(aWinButton)[BUTTON_TOP]         ,,;
        BUTTON                                  ;
        STYLE BUTTON_TYPE_RELEASE               ;
        ACTION { || ScanImg(oBrowseImg) }       ;
        ACCELERATOR "S"                         ;
        DISPLAYUP gBmpLoad("SCAN.BMU")          ;
        DISPLAYDN gBmpLoad("SCAN.BMD")          ;
        ATTACH aWinButtons


                                // -LLIBG- Show all buttons defined
      SHOW ALL BUTTONS aWinButtons


      // Main loop
      WHILE lMore

         IF !oBrowseImg:lIsStable .AND. NEXTKEY()==0

            ImgDynaDisplay(oBrowseImg)

         ENDIF

         IF oBrowseImg:lIsStable .AND. NEXTKEY()==0
                             //         When the browse move, with K_UP,K_DOWN...
                             //         we need to refresh scrollbars
                             //         locations. Refresh only if no key pending
            ScrBarUpDate(aHoriScroll,ImageHoriBar(NIL,NIL,oBrowseImg))
            ScrBarUpDate(aVertScroll,ImageVertBar(NIL,NIL,oBrowseImg))

         ENDIF

         // Everything's done -- just wait for a key
         nKey := mInkey( 0 , aWinButtons )
                             // -LLIBG- Use mInkey() instead of INKEY() and
                             //         add aWinButtons allow buttons management


         IF ( nKey == K_ESC )
            // Esc means leave
            lMore := .F.

         ELSEIF ( nKey == K_BUTTON )   .OR. ;
                ( nKey == K_CLIC_OUT ) .OR. ;
                ( nKey == K_ACCELERATOR )

            // -LLIBG- Warning, when a button is hitted, mInkey() give the hand
            //         back to main calling routine, and a value of K_BUTTON is
            //         returned. We must not apply this K_BUTTON to applyKey
            //         K_CLIC_OUT is returned when the clic does not match
            //         any button. K_ACCELERATOR is returned when an accelerator
            //         key have been striked
            //         We dont need to apply these keys


         ELSE

            // Apply the key to the ImageBrowse
            EVAL( oBrowseImg:bApplyKey, oBrowseImg, nKey )

         ENDIF

      ENDDO

   ENDIF

                             // Erase VMM item
   oBrowseImg:aDisPtr := iErase(oBrowseImg:aDisPtr)
   oBrowseImg:aOriPtr := iErase(oBrowseImg:aOriPtr)

                             // General restoring
   RESTSCREEN( nTop, nLeft, nBottom, nRight, xScreen)
   SETCOLOR(cColorSave)
   SETCURSOR(nCursoSave)

   RETURN (NIL)



