/***
*
*  iFormat.prg
*
*  Illustration of images files formats with Light Lib Images.
*
*  Copyright (c) 1993, DFL.
*  All rights reserved.
*
*/


#include "Inkey.ch"

#include "llibi.ch"          //  Use Light Lib Images defines
#include "llibitoo.ch"       //  Use Light Lib Images tools defines
#include "llibg.ch"          //  Use Light Lib Graphics defines
#include "llibgtoo.ch"       //  Use Light Lib Graphics tools defines

MEMVAR GetList               // To avoid warning when clipper /w


*
FUNCTION iFormat()           // Retrieve all images filenames from current drive

   LOCAL nTop    := 06
   LOCAL nLeft   := 02
   LOCAL nBottom := 24
   LOCAL nRight  := 75
   LOCAL cFile   := ""
   LOCAL nI      := 0

   LOCAL aDir    := {}
   LOCAL aFiles  := {}

   aDir := ASORT(DIRECTORY("*.TIF"),,,{|e1,e2| e1[1]<e2[1]})
   FOR nI := 1 TO LEN(aDir)
       AADD(aFiles,PADR(aDir[nI,1],12)+' - '+TRANSFORM(aDir[nI,2],'999999'))
   NEXT nI

   aDir := ASORT(DIRECTORY("*.BMP"),,,{|e1,e2| e1[1]<e2[1]})
   FOR nI := 1 TO LEN(aDir)
       AADD(aFiles,PADR(aDir[nI,1],12)+' - '+TRANSFORM(aDir[nI,2],'999999'))
   NEXT nI

   aDir := ASORT(DIRECTORY("*.PCX"),,,{|e1,e2| e1[1]<e2[1]})
   FOR nI := 1 TO LEN(aDir)
       AADD(aFiles,PADR(aDir[nI,1],12)+' - '+TRANSFORM(aDir[nI,2],'999999'))
   NEXT nI


   DO WHILE (cFile := xDropBoxBrowse(00,00,MAXROW(),28,aFiles))<>NIL

                             // Execute on selected file
      FileFormat(LEFT(cFile,12))


   ENDDO


   RETURN (NIL)


*
STATIC FUNCTION FileFormat(cFileName ;   // File to be converted
                          )

   LOCAL nTop         := 1   // Windows coordinates
   LOCAL nLeft        := 2   // in rows and columns
   LOCAL nBottom      :=17   //
   LOCAL nRight       :=59   //

   LOCAL xSave        := SAVESCREEN(nTop, nLeft, nBottom, nRight)

   LOCAL cSaveColor   := SETCOLOR("N/W,W+/B,,G,W+/N")

   LOCAL aWinButtons  := {}  // Create an empty button handler

   LOCAL cOriFileName := LEFT(cFileName,AT('.',cFileName)-1)
   LOCAL cOriFileFor  := ''  // Original file format
   LOCAL nImgWidth    := 0   // Original image width
   LOCAL nImgHeight   := 0   // Original image height
   LOCAL nImgColors   := 0   // Original image colors
   LOCAL nImgDensity  := 0   // Original image density
   LOCAL cImgDensity  := ''  // Original image density in carac

                             // Destin image informations
   LOCAL cDesFileName := 'MyFile  '
   LOCAL cDesFileFor  := 'TIF'
   LOCAL lDesFileComp := .T.
   LOCAL cSaveWhat    := 'All '

   LOCAL aImgPtr             // Image ptr
                             // Set saver
   LOCAL aSetSave     := iSet( LLI_SET_UDF , { { | nStatus , nCall | lImgGauge( nStatus , nCall ) } , 10 } )

                             // Display a 3D box
   DISPBOX(nTop, nLeft, nBottom, nRight, LLG_BOX_GRAY_SQUARE)

                             // Depending on extension, use the correct file
                             // format to load the image file
   IF '.BMP'$cFileName

                             // Read .BMP file
      aImgPtr := iGet( LLI_DISK, LLI_DISK_BMP,  0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, { cFileName } )
      cOriFileFor := 'BMP'

   ELSEIF '.PCX'$cFileName

                             // Read .PCX file
      aImgPtr := iGet( LLI_DISK, LLI_DISK_PCX,  0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, { cFileName } )
      cOriFileFor := 'PCX'

   ELSEIF '.TIF'$cFileName

                             // Read .TIF file
      aImgPtr := iGet( LLI_DISK, LLI_DISK_TIF,  0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, { cFileName } )
      cOriFileFor := 'TIF'

   ENDIF


   IF !EMPTY(aImgPtr)

                             // Retrieve values from image structure
      nImgWidth   := aImgPtr[LLI_IMG_WIDTH]
      nImgHeight  := aImgPtr[LLI_IMG_HEIGHT]
      nImgColors  := aImgPtr[LLI_IMG_COLORS]
      nImgDensity := aImgPtr[LLI_IMG_DPI]
      cImgDensity := STR(nImgDensity,3)

                             // Empty the GetList
      GetList := {}

                             // Original image informations
      @ nTop+02 , nLeft+04 SAY "FileName" GET cOriFileName  PICTURE 'XXXXXXXX' WHEN .F.
      @ nTop+03 , nLeft+04 SAY "Width   " GET nImgWidth     PICTURE '9999' INCREMENTAL ATTACH aWinButtons WHEN .F.
      @ nTop+04 , nLeft+04 SAY "Height  " GET nImgHeight    PICTURE '9999' INCREMENTAL ATTACH aWinButtons WHEN .F.
      @ nTop+05 , nLeft+04 SAY "Colors  " GET nImgColors    PICTURE '9999' INCREMENTAL ATTACH aWinButtons WHEN .F.
      GetBox(nTop+02,nLeft+04,nTop+05,nLeft+20,4,'Size')

      @ nTop+02 , nLeft+25 SAY "Density" GET cImgDensity  RADIO " 75","100","150","300" ATTACH aWinButtons
      GetBox(nTop+02,nLeft+25,nTop+05,nLeft+37,4,'Density')

      @ nTop+02 , nLeft+42 SAY "Format" GET cOriFileForm RADIO "TIF","PCX","BMP" ATTACH aWinButtons
      GetBox(nTop+02,nLeft+42,nTop+04,nLeft+53,4,'Format')

      GetBox(nTop+02,nLeft+04,nTop+05,nLeft+53,8,'Original')

                             // Destination file informations
      @ nTop+08 , nLeft+04 SAY "FileName" GET cDesFileName PICTURE 'XXXXXXXX'
      @ nTop+10 , nLeft+11 SAY "Compress" GET lDesFileComp CHECKBOX ATTACH aWinButtons
      GetBox(nTop+08,nLeft+04,nTop+10,nLeft+20,4,'File')

      @ nTop+08 , nLeft+25 SAY "Format "  GET cDesFileForm RADIO "TIF","PCX","BMP" ATTACH aWinButtons
      GetBox(nTop+08,nLeft+25,nTop+10,nLeft+37,4,'Format')

                             // Get what to save informations
      @ nTop+08 , nLeft+42 SAY "What "    GET cSaveWhat    RADIO 'All ','Area' ATTACH aWinButtons
      GetBox(nTop+08,nLeft+42,nTop+09,nLeft+53,4,'Area')

      GetBox(nTop+08,nLeft+04,nTop+10,nLeft+53,8,'To save')



                             // Add a "CANCEL" button to emulate K_ESC
      @ (nLeft+10) * FONT_SIZE_X                ,;
        (nBottom-03) * FONT_SIZE_Y             ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_KEY                    ;
        ACTION K_ESC                             ;
        ACCELERATOR K_ALT_L                      ;
        DISPLAYUP gBmpLoad("CANCEL.BMU")         ;
        DISPLAYDN gBmpLoad("CANCEL.BMD")         ;
        ATTACH aWinButtons

                             // Add a "VIEW" button to view the image
      @ (nLeft+20) * FONT_SIZE_X                ,;
        (nBottom-03) * FONT_SIZE_Y             ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_RELEASE                ;
        ACTION { || ImageViewer(13,02,29,76,cFileName) } ;
        ACCELERATOR K_ALT_V                      ;
        DISPLAYUP gBmpLoad("VIEW.BMU")           ;
        DISPLAYDN gBmpLoad("VIEW.BMD")           ;
        ATTACH aWinButtons

                             // Add a "RUN" button to run the new format
      @ (nLeft+30) * FONT_SIZE_X                ,;
        (nBottom-03) * FONT_SIZE_Y             ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_RELEASE                ;
        ACTION { || ImageSaver(aImgPtr,cDesFileName,lDesFileComp,cDesFileForm,cSaveWhat) }               ;
        ACCELERATOR K_ALT_R                      ;
        DISPLAYUP gBmpLoad("RUN.BMU")            ;
        DISPLAYDN gBmpLoad("RUN.BMD")            ;
        ATTACH aWinButtons

                             // Add a close window button
      @ nLeft*FONT_SIZE_X                       ,;
        nTop *FONT_SIZE_Y                      ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_KEY                    ;
        ACTION K_ESC                             ;
        DISPLAYUP gBmpLoad("CLOSEWIN.BMU")       ;
        DISPLAYDN gBmpLoad("CLOSEWIN.BMD")       ;
        ATTACH aWinButtons

                             // Show all buttons in aWinButtons list
      SHOW ALL BUTTONS aWinButtons

                             // Read gets, starting get 8, managing
                             // aWinButtons buttons
      READ INITGET 12 BUTTONS aWinButtons

   ENDIF

   aImgPtr := iErase(aImgPtr)
   iSet( LLI_SET_UDF , aSetSave )

                             // Restore window
   RESTSCREEN(nTop,nLeft,nBottom,nRight,xSave)
   SETCOLOR(cSaveColor)      // Restore colors


   RETURN (NIL)



*
STATIC FUNCTION ImageSaver(aImgPtr      ,;       // Original image
                           cDesFileName ,;       // Destination filename
                           lDesFileComp ,;       // Need compression
                           cDesFileForm ,;       // Destination fileformat
                           cSaveWhat     ;       // Partial/all saving
                          )


   LOCAL nFileForm := LLI_DISK_TIF
   LOCAL nFileComp := IF(lDesFileComp,LLI_DISK_COMPRESS_AUTO,LLI_DISK_COMPRESS_NIL)
   LOCAL cFileName := ''
   LOCAL aSetSave  := iSet( LLI_SET_UDF , { { | nStatus , nCall | lImgGauge( nStatus , nCall ) } , 10 } )

                             // By default save all the image
   LOCAL nSaveX1   := LLI_FULL_SIZE
   LOCAL nSaveY1   := LLI_FULL_SIZE
   LOCAL nSaveX2   := LLI_FULL_SIZE
   LOCAL nSaveY2   := LLI_FULL_SIZE

                             // Partial selection
   LOCAL nLeft       := 0
   LOCAL nTop        := 0
   LOCAL nRight      := 0
   LOCAL nBottom     := 0
   LOCAL xSaveScreen := NIL
   LOCAL xSaveCursor := NIL
   LOCAL aTmpPtr     := NIL
   LOCAL aSelected   := {}   // Selected coordinates
   LOCAL cTitle      := ''


                             // Area to be saved must be choose here
   IF cSaveWhat == 'Area'

      nTop    := 07
      nLeft   := 13
      nBottom := 22
      nRight  := nLeft + 2 * INT((nBottom-nTop+1)/aImgPtr[LLI_IMG_HEIGHT]*aImgPtr[LLI_IMG_WIDTH])

                             // Save environnement
      xSaveCursor := SETCURSOR(0)
      xSaveScreen := SAVESCREEN(nTop,nLeft,nBottom,nRight)

                             // Display a 3D box
      DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Display a title in box frame
      cTitle := LEFT("Use mouse to select area",nRight-nLeft+1)
      gWriteAt( ( nLeft+INT(((nRight-nLeft+1)-LEN(cTitle))/2 ))* FONT_SIZE_X ,;
                nTop * FONT_SIZE_Y                                           ,;
                cTitle                                                       ,;
                1                                                            ,;
                LLG_MODE_SET                                                  ;
              )


                             // Execute zoom to fit original image
                             // in a small select window
      aTmpPtr := iCopy( aImgPtr                                                      ,;
                        0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE                           ,;
                        LLI_COPY_ZOOM                                                ,;
                        { INT((nRight-nLeft-3)*FONT_SIZE_X)                          ,;
                          INT((nBottom-nTop-1)*FONT_SIZE_Y)                           ;
                        }                                                             ;
                      )


                             // Put small image on screen
      iPut( aTmpPtr                            ,;
            0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE ,;
            LLI_SCREEN, NIL                    ,;
            (nLeft + 2)*FONT_SIZE_X            ,;
            (nTop  + 1)*FONT_SIZE_Y            ,;
            {.T.}                               ;
          )


                             // Select area, starting with mouse at the middle
                             // of the image window
      aSelected := aImgSelArea( (nLeft   + 2)*FONT_SIZE_X ,;
                                (nTop    + 1)*FONT_SIZE_Y ,;
                                (nRight  - 1)*FONT_SIZE_X ,;
                                (nBottom    )*FONT_SIZE_Y  ;
                              )


      IF aSelected[AREA_SELECTED]
                             // If area have been selected

         nSaveX1 := INT((aSelected[AREA_X1]-(nLeft+2)*FONT_SIZE_X)/(INT((nRight-nLeft-3)*FONT_SIZE_X)/aImgPtr[LLI_IMG_WIDTH]))
         nSaveY1 := INT((aSelected[AREA_Y1]-(nTop +1)*FONT_SIZE_Y)/(INT((nBottom-nTop-1)*FONT_SIZE_Y)/aImgPtr[LLI_IMG_HEIGHT]))
         nSaveX2 := INT((aSelected[AREA_X2]-(nLeft+2)*FONT_SIZE_X)/(INT((nRight-nLeft-3)*FONT_SIZE_X)/aImgPtr[LLI_IMG_WIDTH]))
         nSaveY2 := INT((aSelected[AREA_Y2]-(nTop +1)*FONT_SIZE_Y)/(INT((nBottom-nTop-1)*FONT_SIZE_Y)/aImgPtr[LLI_IMG_HEIGHT]))

      ELSE
                              // If no aera selected, use full size
         nSaveX1 := LLI_FULL_SIZE
         nSaveY1 := LLI_FULL_SIZE
         nSaveX2 := LLI_FULL_SIZE
         nSaveY2 := LLI_FULL_SIZE

      ENDIF

                             // Erase temporary image
      aTmpPtr := iErase(aTmpPtr)

                             // Restore window
      RESTSCREEN(nTop,nLeft,nBottom,nRight,xSaveScreen)
      SETCURSOR(xSaveCursor)

   ENDIF

                             // Determine file format
   IF cDesFileForm == "TIF"

      cFileName := ALLTRIM(cDesFileName)+'.TIF'
      nFileForm := LLI_DISK_TIF

   ELSEIF cDesFileForm == "BMP"

      cFileName := ALLTRIM(cDesFileName)+'.BMP'
      nFileForm := LLI_DISK_BMP

   ELSEIF cDesFileForm == "PCX"

      cFileName := ALLTRIM(cDesFileName)+'.PCX'
      nFileForm := LLI_DISK_PCX

   ENDIF

                             // Save the image in filename with good fileformat
                             // and compression mode
   iPut( aImgPtr                      ,;
         nSaveX1                      ,;
         nSaveY1                      ,;
         nSaveX2                      ,;
         nSaveY2                      ,;
         LLI_DISK, nFileForm          ,;
         0 , 0                        ,;
         { cFileName  , nFileComp }    ;
      )

   iSet( LLI_SET_UDF , aSetSave )


   RETURN (NIL)


*
STATIC FUNCTION lImgGauge(nCallMode     ,;   // Call mose ie LLI_UDF_INIT | IDLE ...
                          nValue         ;   // Value from Light Lib Images
                         )

                             // Display a gauge and follow images
                             // operations

   STATIC aHandle   := {}    // Handle on gauge

                             // By default, continue operations
   LOCAL  lContinue := LLI_UDF_CONT


   IF nCallMode == LLI_UDF_INIT

                             // Create a new gauge and
                             // set the number of futur calls
      lGauge(GAUGE_INIT             ,;
             aHandle                ,;
             GAUGE_VERT             ,;
             0                      ,;
             MAXCOL()-1             ,;
             MAXROW()+1             ,;
             { || nValue }          ,;  // Max value
             { || nGaugeCurValue() } ;
            )

   ELSEIF nCallMode == LLI_UDF_IDLE

                             // Set current value
      nGaugeCurValue(nValue)
                             // Display Gauge
      lGauge(GAUGE_IDLE,aHandle)

      IF NEXTKEY()<>0
         lContinue := LLI_UDF_ABORT
      ENDIF

   ELSEIF nCallMode == LLI_UDF_EXIT

                             // Close Gauge
      lGauge(GAUGE_EXIT,aHandle)

   ELSEIF nCallMode == LLI_UDF_ERROR

      nMessageBox("Read, write or display error" ,;
                  "Error"                        ,;
                  MB_OK                          ,;
                  MB_ICON_STOP                    ;
               )

   ENDIF


   RETURN (lContinue)

*
STATIC FUNCTION nGaugeCurValue(nValue)
                             // Set Get function to allow Gauge use

   STATIC nCurValue := 0

   IF nValue<>NIL

      nCurValue := nValue

   ENDIF

   RETURN (nCurValue)


