/***
*
*  IVirtual.prg
*
*  Illustration of virtual terminal with Light Lib Images.
*
*  Copyright (c) 1993, DFL.
*  All rights reserved.
*
*/


#include "Inkey.ch"

#include "llibi.ch"          //  Use Light Lib Images defines
#include "llibitoo.ch"       //  Use Light Lib Images tools defines
#include "llibg.ch"          //  Use Light Lib Graphics defines
#include "llibgtoo.ch"       //  Use Light Lib Graphics tools defines



*
FUNCTION VirtualTerminal( nTop, nleft, nBottom, nRight)


   LOCAL nKey                // Keystroke
   LOCAL lMore := .T.        // Loop control

   LOCAL aWinButtons  := {}  //  Arrays to handle buttons
   LOCAL aVertScroll  := {}  //  and scrolls bars
   LOCAL aHoriScroll  := {}  //

                             // General savings
   LOCAL xScreen     := SAVESCREEN( nTop, nLeft, nBottom, nRight)
   LOCAL cColorSave  := SETCOLOR('N/W,W+/G,W,W,GR+/W')
   LOCAL nCursoSave  := SETCURSOR(0)
   LOCAL xScreenTemp := ""


   LOCAL nMaxCol := 0
   LOCAL nMaxRow := 0

                             // Browse pseudo object
   LOCAL oBrowseImg  := ARRAY(BROWSE_IMG_SIZE)

   oBrowseImg:lIsStable      := .F.
   oBrowseImg:nCurrentX      := 0
   oBrowseImg:nCurrentY      := 0
   oBrowseImg:nScaleX        := 0
   oBrowseImg:nScaleY        := 0

                             // No image loaded, ptr are empty
   oBrowseImg:aOriPtr        := {}
   oBrowseImg:aDisPtr        := {}

                             // Compute display window pixels values
   oBrowseImg:nWinDispX1     := FONT_SIZE_X * ( nLeft + 2 )
   oBrowseImg:nWinDispY1     := FONT_SIZE_Y * ( nTop + 3 )
   oBrowseImg:nWinDispX2     := FONT_SIZE_X * ( nRight - 3 )  - 1
   oBrowseImg:nWinDispY2     := FONT_SIZE_Y * ( nBottom - 1 ) - 1
   oBrowseImg:nWinDispWidth  := oBrowseImg:nWinDispX2 - oBrowseImg:nWinDispX1 + 1
   oBrowseImg:nWinDispHeight := oBrowseImg:nWinDispY2 - oBrowseImg:nWinDispY1 + 1

                             // Fill the images to be browsed
   oBrowseImg:aImgAvail      := {}
   oBrowseImg:nImgInUse      := 0

                             // Set the block to standard apply key
   oBrowseImg:bApplyKey      := { |oBrowse,nKey| ImgApplyKey( oBrowse, nKey ) }
   oBrowseImg:aUdfShort      := { { | nStatus , nCall | lImgGauge( nStatus , nCall, .F., oBrowseImg ) } , 10 }
   oBrowseImg:aUdfLong       := { { | nStatus , nCall | lImgGauge( nStatus , nCall, .T., oBrowseImg ) } , 20 }

                             // -LLIBG- Display 3D box
   DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Set udf to long operation display
   iSet( LLI_SET_UDF , oBrowseImg:aUdfLong )

   IF gMode()[LLG_MODE_COLOR_MAX]==256
                             // Create a 256 colors images in the memory,
                             // 1000 x 1000 pixels, fill with color 3
      oBrowseImg:aOriPtr := iGet( LLI_MEMORY, LLI_MEMORY_256,  0, 0, 999, 999, { 7 } )

   ELSE
                             // Create a 16 colors images in the memory,
                             // 1000 x 1000 pixels, fill with color 3
      oBrowseImg:aOriPtr := iGet( LLI_MEMORY, LLI_MEMORY_16,  0, 0, 999, 999, { 7 } )

   ENDIF

                             // Save the screen as DisplaySomething() function
                             // will alter it
   xScreenTemp := SAVESCREEN( 0, 0, MAXROW(), MAXCOL() )

   CLS                       // Clear the screen
   DisplaySomething()        // Display some samples on classical screen

                             // Switch terminal to virtual image
   iSet( LLI_SET_TERM, { oBrowseImg:aOriPtr } )

   DisplaySomething()        // Display the same samples in the virtual screen

                             // Switch back terminal to screen
   iSet( LLI_SET_TERM, LLI_VIDEO_SCREEN )

                             // Wait for user to strike a key
   @ MAXROW(),0 SAY "Strike any key to continue"
   INKEY(0)

                             // Restore old screen
   RESTSCREEN( 0, 0, MAXROW(), MAXCOL(),xScreenTemp )

                             // Now allow the user to browse the virtual
                             // screen previouly filled with something

                             // Duplicate image to allow image to be scaled
   oBrowseImg:aDisPtr := iCopy( oBrowseImg:aOriPtr, 0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, LLI_COPY_CLONE )

                             // Set gauge function
   iSet( LLI_SET_UDF , oBrowseImg:aUdfShort )


                             // Add vertical scroll bar
   @ (nRight-3)*FONT_SIZE_X                                             ,;
     (nTop+1)*FONT_SIZE_Y                                               ,;
     (nRight-1)*FONT_SIZE_X                                             ,;
     (nBottom-1)*FONT_SIZE_Y                                             ;
     SCROLLBAR                                                           ;
     STYLE BUTTON_TYPE_SCROLL_VERT                                       ;
     ACTIONUP  { || NIL }                                                ;
     ACTIONDN  { || NIL }                                                ;
     ACTIONBAR { |nSens,nPercent| ImageVertBar(nSens,nPercent,oBrowseImg) }         ;
     HANDLE aVertScroll                                                  ;
     ATTACH aWinButtons


                             // -LLIBG- Add horizontal scroll bar
   @ (nLeft+2)*FONT_SIZE_X                                              ,;
     (nBottom-1)*FONT_SIZE_Y                                            ,;
     (nRight-3)*FONT_SIZE_X                                             ,;
     (nBottom-0)*FONT_SIZE_Y                                             ;
     SCROLLBAR                                                           ;
     STYLE BUTTON_TYPE_SCROLL_HORI                                       ;
     ACTIONUP  { || NIL }                                                ;
     ACTIONDN  { || NIL }                                                ;
     ACTIONBAR { |nSens,nPercent| ImageHoriBar(nSens,nPercent,oBrowseImg) }         ;
     HANDLE aHoriScroll                                                  ;
     ATTACH aWinButtons

                             // Add a close window button
   @ nLeft*FONT_SIZE_X                       ,;
     nTop *FONT_SIZE_Y                      ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_KEY                    ;
     ACTION K_ESC                             ;
     DISPLAYUP gBmpLoad("CLOSEWIN.BMU")       ;
     DISPLAYDN gBmpLoad("CLOSEWIN.BMD")       ;
     ATTACH aWinButtons

                             // -LLIBG- Add an icon bar


                             // -LLIBG- Add a button to allow user
                             //         to quit
   @ (nLeft+2 )*FONT_SIZE_X                  ,;
     (nTop+1  )*FONT_SIZE_Y                 ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_KEY                    ;
     ACTION K_ESC                             ;
     ACCELERATOR "Q"                          ;
     DISPLAYUP gBmpLoad("QUIT.BMU")           ;
     DISPLAYDN gBmpLoad("QUIT.BMD")           ;
     ATTACH aWinButtons


                             // -LLIBG- Add a button to fit image into
                             //         the window
   @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
     ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
     BUTTON                                                             ;
     STYLE BUTTON_TYPE_RELEASE                                          ;
     ACTION { || ImgZoom(ASC('/'),oBrowseImg) }                         ;
     ACCELERATOR "S"                                                    ;
     DISPLAYUP gBmpLoad("SMALEST.BMU")                                  ;
     DISPLAYDN gBmpLoad("SMALEST.BMD")                                  ;
     ATTACH aWinButtons

                             // -LLIBG- Add a button to zoom a part of
                             //         the image
   @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
     ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
     BUTTON                                                             ;
     STYLE BUTTON_TYPE_RELEASE                                          ;
     ACTION { || ImgZoom(ASC(''),oBrowseImg) }                         ;
     ACCELERATOR "A"                                                    ;
     DISPLAYUP gBmpLoad("AREA.BMU")                                     ;
     DISPLAYDN gBmpLoad("AREA.BMD")                                     ;
     ATTACH aWinButtons


                             // -LLIBG- Add a button to rotate to the right
                             //         the image
   @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
     ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
     BUTTON                                                             ;
     STYLE BUTTON_TYPE_RELEASE                                          ;
     ACTION { || ImgRotate(90,oBrowseImg) }                             ;
     ACCELERATOR "9"                                                    ;
     DISPLAYUP gBmpLoad("ROTAT090.BMU")                                 ;
     DISPLAYDN gBmpLoad("ROTAT090.BMD")                                 ;
     ATTACH aWinButtons

                             // -LLIBG- Add a button to flip the image upside
                             //         down
   @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
     ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
     BUTTON                                                             ;
     STYLE BUTTON_TYPE_RELEASE                                          ;
     ACTION { || ImgRotate(180,oBrowseImg) }                            ;
     ACCELERATOR "1"                                                    ;
     DISPLAYUP gBmpLoad("ROTAT180.BMU")                                 ;
     DISPLAYDN gBmpLoad("ROTAT180.BMD")                                 ;
     ATTACH aWinButtons


                             // -LLIBG- Add a button to allow printing
   @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
     ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
     BUTTON                                                             ;
     STYLE BUTTON_TYPE_RELEASE                                          ;
     ACTION { || PrintImg(oBrowseImg) }                                 ;
     ACCELERATOR "P"                                                    ;
     DISPLAYUP gBmpLoad("PRINT.BMU")                                    ;
     DISPLAYDN gBmpLoad("PRINT.BMD")                                    ;
     ATTACH aWinButtons


                             // -LLIBG- Show all buttons defined
   SHOW ALL BUTTONS aWinButtons


   // Main loop
   WHILE lMore

      IF !oBrowseImg:lIsStable .AND. NEXTKEY()==0

         ImgDynaDisplay(oBrowseImg)

      ENDIF

      IF oBrowseImg:lIsStable .AND. NEXTKEY()==0
                             // When the browse move, with K_UP,K_DOWN...
                             // we need to refresh scrollbars
                             // locations. Refresh only if no key pending
         ScrBarUpDate(aHoriScroll,ImageHoriBar(NIL,NIL,oBrowseImg))
         ScrBarUpDate(aVertScroll,ImageVertBar(NIL,NIL,oBrowseImg))

      ENDIF

                             // Everything's done -- just wait for a key
      nKey := mInkey( 0 , aWinButtons )
                             // Use mInkey() instead of INKEY() and
                             // add aWinButtons allow buttons management


      IF ( nKey == K_ESC )
                             // Esc means leave
         lMore := .F.

      ELSEIF ( nKey == K_BUTTON )   .OR. ;
             ( nKey == K_CLIC_OUT ) .OR. ;
             ( nKey == K_ACCELERATOR )

         //  Warning, when a button is hitted, mInkey() give the hand
         //  back to main calling routine, and a value of K_BUTTON is
         //  returned. We must not apply this K_BUTTON to applyKey
         //  K_CLIC_OUT is returned when the clic does not match
         //  any button. K_ACCELERATOR is returned when an accelerator
         //  key have been striked
         //  We dont need to apply these keys

      ELSE

                             // Apply the key to the ImageBrowse
         EVAL( oBrowseImg:bApplyKey, oBrowseImg, nKey )

      ENDIF

   ENDDO
                             // Erase VMM item
   oBrowseImg:aDisPtr := iErase(oBrowseImg:aDisPtr)
   oBrowseImg:aOriPtr := iErase(oBrowseImg:aOriPtr)

                             // General restoring
   RESTSCREEN( nTop, nLeft, nBottom, nRight, xScreen)
   SETCOLOR(cColorSave)
   SETCURSOR(nCursoSave)

   RETURN (NIL)

*
STATIC FUNCTION DisplaySomething()

                             // Just display something on the screen using
                             // CA-CLIPPER or Light Lib Graphics functions

   LOCAL aValues := aBuildValues(10,3)

                             // Draw a rectangle
   gRect(   0,   0, 100, 100, LLG_FILL ,  8, LLG_MODE_SET )

                             // Draw a line
   gLine( 0, 0, 100, 100, 11, LLG_MODE_SET )

                             // Draw a framed rectangle
   gRect( 110, 110, 200, 200, LLG_FRAME,  5, LLG_MODE_SET )

                             // Draw a filled rectangle
   gRect( 120, 120, 190, 190, LLG_FILL ,  1, LLG_MODE_SET )

                             // Draw a filled ellipse
   gEllipse( 200, 370, 100, 100, NIL, NIL, LLG_FILL, 12 , LLG_MODE_SET)

                             // Draw a frame
   gFrame( 250, 250, 350, 350, 7, 15, 8, 3, 3, 3, 3 )

                             // Write at pixel location
   gWriteAt( 10, 10, "Hello World !", 9  )

                             // Write unsing standard CA-CLIPPER devout()
   @ 10,10 SAY "Hello World" COLOR "W/GR"

                             // Display a box
   DispBox( 10, 10, 16, 30, 2, "GR+/B" )

                             // Display MaxRow MaxCol values
   @ MAXROW()-5,0 SAY "MaxCol   :"+STR(MAXCOL(),4)
   @ MAXROW()-4,0 SAY "MaxRow   :"+STR(MAXROW(),4)
                             // Display MaxRow MaxCol values
   @ MAXROW()-3,0 SAY "Pixels X :"+STR(gMode()[LLG_MODE_GRAPH_COL],4)
   @ MAXROW()-2,0 SAY "Pixels Y :"+STR(gMode()[LLG_MODE_GRAPH_ROW],4)

                             // Display a Bar graph
   BarGraph( MAXROW()-20, MAXCOL()-40, MAXROW()-4, MAXCOL(), aValues, STYLE_BAR )


   RETURN (NIL)

*
STATIC FUNCTION aBuildValues(nNbVal  ,;  // Number of values
                             nMaxCol  ;  // Number of columns
                            )

                             // This function is used to retrieve some values
                             // from DBF file or create them

   LOCAL aReturn  := {}      // Array of values
   LOCAL nI       := 0       // Loop pointer
   LOCAL nJ       := 0       // Loop pointer
   LOCAL nOldRec  := 0       // Current RECNO()
   LOCAL aNum     := {}      // Array

                             // If a file is open
   IF !EMPTY(ALIAS())
                             // Look for numeric fields
      FOR nI := 1 TO FCOUNT()
         IF VALTYPE(FIELDGET(nI)) == 'N'
            AADD(aNum,nI)
         ENDIF
         IF LEN(aNum) > nMaxCol
            EXIT
         ENDIF
      NEXT nI

      nOldRec := RECNO()     // Keep current record number

      FOR nI := 1 TO nNbVal  // Load nNbVal records values
         IF EOF()
            EXIT
         ENDIF
                             // First field will be X axis legend
         AADD(aReturn, { TRANSFORM( FIELDGET(1),'@X') } )

         FOR nJ := 1 TO LEN(aNum)
            AADD(aReturn[LEN(aReturn)], FIELDGET(aNum[nJ]) )
         NEXT nJ
         SKIP
      NEXT
                             // Return to previous record
      GOTO nOldRec
   ENDIF
                             // If file neither has records nor numeric fields
                             // build an array of sequential values
   IF EMPTY(aReturn)
      FOR nI := 1 TO nNbVal
         AADD(aReturn, { REPLICATE( CHR( nI + 64),3) , nI, INT(nI/LOG(nI)) } )
      NEXT
   ENDIF

   RETURN (aReturn)
/*

*
FUNCTION BarGrap1( nTop     ,;  // Top coordinates in rows
                   nleft    ,;  // Left coordinates in columns
                   nBottom  ,;  // Bottom coordinates in rows
                   nRight   ,;  // Left coordinates in columns
                   aValues  ,;  // List of values
                   nStyle   ,;  // Graph style (STYLE_BAR, STYLE_LINE, STYLE_BAR_3D, STYLE_LINE_3D)
                   nMin     ,;  // Vertical scale minimum value
                   nMax      ;  // Vertical scale maximum value
                 )



   LOCAL aOldClip     := {}                                // Previous Clipping region

   LOCAL nI           := 0                                 // Loop pointer
   LOCAL nJ           := 0                                 // Loop pointer
   LOCAL nIncrement   := 0                                 // Space beetween each value
   LOCAL nMaxValue    := 0                                 // Maximum value of data set
   LOCAL nMinValue    := 0                                 // Minimum value of data set
   LOCAL nNbBar       := 0                                 // Number of bars for each value
   Local nBarWidth    := 0                                 // Bar width

   LOCAL nGraphTop    := (nTop * FONT_SIZE_Y + 20)         // Top   limit of graph area
   LOCAL nGraphleft   := (nLeft * FONT_SIZE_X + 16 +10 )   // Left     ""
   LOCAL nGraphBottom := (nBottom * FONT_SIZE_Y - 20)      // Bottom   ""
   LOCAL nGraphRight  := (nRight * FONT_SIZE_X - 16 -10)   // Right    ""

   LOCAL nXPosition   := nGraphBottom                     // Position of X axis
   LOCAL nXAxisColor  := 4                                // Color of X axis
   LOCAL nYPosition   := nGraphLeft                       // Position of Y axis
   LOCAL nYAxisColor  := 3                                // Color of Y axis

   LOCAL nGraphWidth  := nGraphRight - nGraphLeft         // Graph width
   LOCAL nGraphHeight := nGraphBottom - nGraphTop         // Graph height

   LOCAL nYlegLen     := 0                                // Length of Y axis legend
   LOCAL nXLegLen     := 0                                // Total Length of X axis legend
   LOCAL nXLegNbLine  := 0                                // Number of lines of X axis legend



                             // Compute Maximum and Minimum values
   FOR nI := 1 TO LEN(aValues)
      FOR nJ := 2 TO LEN(aValues[nI])
         nMaxValue := MAX( aValues[nI,nJ], nMaxValue )
         nMinValue := MIN( aValues[nI,nJ], nMinValue )
      NEXT
                             // Compute maximum number of bars
      nNbBar := MAX(nNbBar,LEN(aValues[nI])-1)

                             // Total lentgh of X axis legends
      nXLegLen := nXLegLen + LEN(aValues[nI,1]) + 1

   NEXT

                             // Round minimum and maximum values
   nMaxValue := nRoundSup(nMaxValue)
   nMinValue := nRoundSup(nMinValue)


                             // If maximum and minimum values have been
                             // defined replace nMinValue and nMaxValue
   IF nMin <> NIL
      nMinValue := nMin
   ENDIF
   IF nMax <> NIL
      nMaxValue := nMax
   ENDIF

                             // Length of Y axis legend is max of
                             // minimum value length and maximum value length
   nYLegLen  := MAX( LEN(ALLTRIM(STR(nMaxValue))), ;
                     LEN(ALLTRIM(STR(nMinValue))))

                             // Y axis position
   nYPosition  := nGraphLeft + nYLegLen * FONT_SIZE_X

                             // Maximum graph width
   nGraphWidth := nGraphRight - nYPosition

                             // Increment to move bars
   nIncrement  := nGraphWidth / LEN(aValues)

                             // Bars width
   nBarWidth   := nIncrement / nNbBar

                             // Number of X axis legend lines
   nXLegNbLine  := INT( nXLegLen * FONT_SIZE_X / nGraphWidth )
   IF nXLegNbLine < nXLegLen * FONT_SIZE_X / nGraphWidth
      nXLegNbLine  := nXLegNbLine  + 1
   ENDIF
                             // Correct graph Height
   nGraphHeight := nGraphHeight - nXLegNbLine * FONT_SIZE_Y

                             // Correct graph bottom position
   nGraphBottom := nGraphBottom - nXLegNbLine * FONT_SIZE_Y

                             // X axis position
   nXPosition  := nGraphBottom - ABS( nMinValue / (nMaxValue-nMinValue) ) * nGraphHeight

                             // Define clipping region to avoid
                             // writing outside graph
   aOldClip := gSetClip( nLeft   * FONT_SIZE_X  ,;
                         nTop    * FONT_SIZE_Y  ,;
                         nRight  * FONT_SIZE_X  ,;
                         nBottom * FONT_SIZE_Y   ;
                       )

                             // Draw box arround graph
   gFrame( nLeft   * FONT_SIZE_X     ,;
           nTop    * FONT_SIZE_Y     ,;
           nRight  * FONT_SIZE_X - 1 ,;
           nBottom * FONT_SIZE_Y - 1 ,;
           7, 15, 8                  ,;
           5, 5, 5, 5                ,;
           LLG_MODE_SET               ;
         )

                             // Draw horizontal axis
   gLine( nYPosition         ,;
          nXPosition         ,;
          nGraphRight        ,;
          nXPosition         ,;
          nXAxisColor        ,;
          LLG_MODE_SET        ;
        )

                             // Draw vertical axis
   gLine( nYPosition         ,;
          nGraphTop          ,;
          nYPosition         ,;
          nGraphBottom       ,;
          nYAxisColor        ,;
          LLG_MODE_SET        ;
        )

                             // Place 10 standard values on Y Axis
   FOR nI := nMinValue TO nMaxValue STEP ABS( (nMinValue-nMaxValue) / 10)
      gWriteAt( nGraphLeft - FONT_SIZE_X                                                 ,;
                nXPosition - nI / ABS( nMinValue - nMaxValue)*nGraphHeight -FONT_SIZE_X  ,;
                STR(nI,nYLegLen) + "-"                                                   ,;
                nYAxisColor                                                              ,;
                LLG_MODE_SET                                                              ;
              )
   NEXT nI


   DrawValu1()               // Reset static variables
                             // keeping previous values

                             // Draw values
   FOR nI := 0 TO LEN(aValues) - 1

                             // Draw bars or a lines for each X value
      FOR nJ := 0 TO nNbBar-1
         DrawValu1( nYPosition + ( nI * nIncrement ) +                             ;
                     IF( nStyle == STYLE_BAR .OR. nStyle == STYLE_BAR_3D                  ,;
                         nJ*nBarWidth                                             ,;
                         0                                                         ;
                       )                                                          ,;
                    nXPosition                                                    ,;
                    nBarWidth                                                     ,;
                    aValues[nI+1,nJ+2] / ABS( nMinValue - nMaxValue)*nGraphHeight ,;
                    nJ+2                                                          ,;
                    nStyle                                                         ;
                  )
      NEXT nJ

                             // Write legends on X axis only for first pass
      gWriteAt(nYPosition + nI * nIncrement                          ,;
               nGraphBottom + (nI % nXLegNbLine ) * FONT_SIZE_Y + 2  ,;
               aValues[nI+1,1]                                       ,;
               nXAxisColor                                           ,;
               LLG_MODE_SET                                           ;
              )

      gLine( nYPosition + nI * nIncrement   ,;
             nXPosition                     ,;
             nYPosition + nI * nIncrement   ,;
             nXPosition + 3                 ,;
             nXAxisColor                    ,;
             LLG_MODE_SET                    ;
           )

   NEXT nI

                             // Reset previous clipping area
   gSetClip( aOldClip[1] ,;
             aOldClip[2] ,;
             aOldClip[3] ,;
             aOldClip[4]  ;
           )

RETURN (NIL)


*
FUNCTION DrawValu1( nLeft    ,; // Left coordinates in pixels
                    nTop     ,; // Top coordinates in pixels
                    nWidth   ,; // Width in pixels
                    nHeight  ,; // Height in pixels
                    nColor   ,; // Color
                    nStyle    ; // Style
                  )

   STATIC aPrevCoord := {}   // Coordinates of previous drawn values


   LOCAL nI       := 0       // Loop pointer
   LOCAL nPointer := 0       // ASCAN pointer
   LOCAL nDepth   := 0       // Depth of 3D graphs

   IF nStyle == NIL          // No parameters => reset aPrevCoord
      nStyle := 99
   ELSEIF nStyle == STYLE_BAR_3D .OR. nStyle == STYLE_LINE_3D
                             // 3D graph, define depth
      nDepth := MIN(nWidth/4,10)
   ELSE
                             // depth = 0
      nDepth := 0
   ENDIF

                             // Style BARGRAPH
   IF nStyle == STYLE_BAR .OR. nStyle == STYLE_BAR_3D

                             // Depth no null 3D bars
      IF nDepth > 0
                             // Fill background whith gray for 3D bar
         FOR nI := 1 TO nDepth

            gRect( nLeft + nI              ,;
                   nTop - nI               ,;
                   nLeft + nWidth + nI     ,;
                   nTop - nHeight - nI     ,;
                   LLG_FRAME               ,;
                   8                       ,;
                   LLG_MODE_SET             ;
                 )

         NEXT nI
                             // Draw 3D bar depth lines
      gLine( nLeft                   ,;
             nTop                    ,;
             nLeft + nDepth          ,;
             nTop - nDepth           ,;
             4                       ,;
             LLG_MODE_SET             ;
           )

      gLine( nLeft                   ,;
             nTop - nHeight          ,;
             nLeft + nDepth          ,;
             nTop - nHeight - nDepth ,;
             4                       ,;
             LLG_MODE_SET             ;
           )

      gLine( nLeft + nWidth          ,;
             nTop                    ,;
             nLeft + nWidth + nDepth ,;
             nTop - nDepth           ,;
             4                       ,;
             LLG_MODE_SET             ;
           )

      gLine( nLeft + nWidth          ,;
             nTop - nHeight          ,;
             nLeft + nWidth + nDepth ,;
             nTop - nHeight - nDepth ,;
             4                       ,;
             LLG_MODE_SET             ;
           )


                             // Draw frame arround background
      gRect( nLeft + nDepth           ,;
             nTop - nDepth            ,;
             nLeft + nWidth + nDepth  ,;
             nTop - nHeight - nDepth  ,;
             LLG_FRAME                ,;
             4                        ,;
             LLG_MODE_SET              ;
           )

   ENDIF
                             // Draw front colored bar
   gRect( nLeft                       ,;
          nTop                        ,;
          nLeft + nWidth              ,;
          nTop - nHeight              ,;
          LLG_FILL                    ,;
          nColor                      ,;
          LLG_MODE_SET                 ;
        )

                             // Draw frame arround colored bar
   gRect( nLeft                       ,;
          nTop                        ,;
          nLeft + nWidth              ,;
          nTop - nHeight              ,;
          LLG_FRAME                   ,;
          4                           ,;
          LLG_MODE_SET                 ;
        )

                             // Style LINEGRAPH
   ELSEIF nStyle == STYLE_LINE .OR. nStyle == STYLE_LINE_3D

      IF nDepth == 0
                             // Draw cross arround point location
                             // with two lines
         gLine( nLeft - 2                ,;
                nTop - nHeight-2         ,;
                nLeft + 2                ,;
                nTop  - nHeight + 2      ,;
                nColor                   ,;
                LLG_MODE_SET              ;
              )

          gLine( nLeft + 2                ,;
                 nTop - nHeight - 2       ,;
                 nLeft - 2                ,;
                 nTop - nHeight + 2       ,;
                 nColor                   ,;
                 LLG_MODE_SET              ;
               )

       ENDIF



                             // Locate coordinates of previous point
                             // with same color
       nPointer  := ASCAN(aPrevCoord, { |aElem| aElem[1] == nColor  })

                             // Previous point found
       IF nPointer  <> 0
          FOR nI := 0 TO nDepth
                             // Draw line between previous point
                             // and current point
             gLine( aPrevCoord[nPointer ,2]+nI  ,;
                    aPrevCoord[nPointer ,3]-nI  ,;
                    nLeft + nI                  ,;
                    nTop - nHeight - nI         ,;
                    nColor                      ,;
                    LLG_MODE_SET                 ;
                  )

          NEXT nI

       ELSE                  // Previous point not found
                             // (first point)
                             // Add an element to coordinates array
                             // for current color
          AADD(aPrevCoord,{nColor,NIL,NIL})
          nPointer  := LEN(aPrevCoord)

       ENDIF
                             // Store current point coordinates for next point
       aPrevCoord[nPointer ,2] := nLeft
       aPrevCoord[nPointer ,3] := nTop - nHeight

    ELSE                     // Style not bar, not line
                             // reset coordinates array
       aPrevCoord := {}

    ENDIF

    RETURN (NIL)

*
STATIC FUNCTION nRoundSup( nValue )


   LOCAL nReturn  := 0       // Return value

                             // Sign of value
   LOCAL nSign    := INT( nValue/ABS(nValue) )

                             // Length of value
   LOCAL nValLen  := LEN( ALLTRIM( STR( INT( nSign * nValue ) ) ) )

   LOCAL nTmpVal  := 0       // Temp Value
   LOCAL nNearest := 0       // Nearest rounded value of nValue


                             // First rounds value
                             // 855 => 900
                             // 820 => 800
   nNearest := ROUND( nSign * nValue, -nValLen + 1)

                             // Keep nearest value
   nReturn  := nNearest

   nTmpVal := nValue

   DO WHILE nSign * nValue > nReturn
                             // If value > rounded value,
                             // then try rounding a greater value
                             // Loop #1    820 * 820 / 800 = 840.5 => 800
                             // Loop #2    840 * 840 / 800 = 882   => 900
      nTmpVal := nTmpVal*nTmpVal / nNearest
      nReturn := ROUND(nTmpVal,-nValLen+1 )
   ENDDO

   RETURN (nSign * nReturn)
*/
