/***
*  LLIBGTOO.CH
*  Tools for Light Lib Graphics definitions
*
*  Copyright (c) 1992-1993, D.F.L.
*  All rights reserved.
*/


//                       ------------- Best video modes definitions -----------

#DEFINE LLG_VIDEO_BEST_16              -16
#DEFINE LLG_VIDEO_BEST_256            -256

//                       ------------- Buttons definitions --------------------

// Button handling array
#DEFINE BUTTON_LEFT                  1    // button left
#DEFINE BUTTON_TOP                   2    // button top
#DEFINE BUTTON_RIGHT                 3    // button right
#DEFINE BUTTON_BOTTOM                4    // button bottom
#DEFINE BUTTON_WIDTH                 5    // button width
#DEFINE BUTTON_HEIGHT                6    // button height
#DEFINE BUTTON_TYPE                  7    // button type
#DEFINE BUTTON_ACTION                8    // button action block -> execute, num -> answer as key
#DEFINE BUTTON_ICO_UP                9    // button icone Up position
#DEFINE BUTTON_ICO_DOWN             10    // button icone Down position
#DEFINE BUTTON_ACCELERATOR          11    // button accelerator Key
#DEFINE BUTTON_CARGO                12    // button cargo value

// Button types
#DEFINE BUTTON_TYPE_KEY              1    // return a special inkey value
#DEFINE BUTTON_TYPE_RELEASE          2    // execute when clic/release
#DEFINE BUTTON_TYPE_REPEAT           3    // execute while clic/down
#DEFINE BUTTON_TYPE_SCROLL_VERT      4    // vertical scroll bar
#DEFINE BUTTON_TYPE_SCROLL_HORI      5    // horizontal scroll bar

//                       ------------- Standard fonts definitions -------------
#DEFINE FONT_SIZE_X                      8    // X size of standard font
#DEFINE FONT_SIZE_Y                     16    // Y size of standard font

//                       ------------- mInkey() return value when an event ----
//                                              had occured
#DEFINE K_BUTTON                     -9999    // Button hitted
#DEFINE K_GET                        -9998    // Get Hitted
#DEFINE K_CLIC_OUT                   -9997    // Clic out of buttons
#DEFINE K_ACCELERATOR                -9996    // Accelerator key hitted


//                       ------------- Standard gauges definitions ------------
#DEFINE GAUGE_INIT               1  // Init phase for gauge function
#DEFINE GAUGE_IDLE             NIL  // Idle phase for gauge function
#DEFINE GAUGE_EXIT               3  // Exit phase for gauge function
#DEFINE GAUGE_VERT               4  // Vertical gauge
#DEFINE GAUGE_HORI               5  // Horizontal gauge

//                       ------------- Standard odometers definitions ---------
#DEFINE ODOME_INIT               1  // Init phase for runstatus function
#DEFINE ODOME_IDLE             NIL  // Idle phase for runstatus function
#DEFINE ODOME_EXIT               3  // Exit phase for runstatus function

//                       ------------- Standard messages box definitions ------
#DEFINE MB_OK                    0  // Message Box with button  OK
#DEFINE MB_OK_CANCEL             1  // Message Box with buttons OK|CANCEL
#DEFINE MB_ABORT_RETRY_IGNORE    2  // Message Box with buttons ABORT|RETRY|IGNORE
#DEFINE MB_YES_NO_CANCEL         3  // Message Box with buttons YES|NO|CANCEL
#DEFINE MB_YES_NO                4  // Message Box with buttons YES|NO
#DEFINE MB_RETRY_CANCEL          5  // Message Box with buttons RETRY|CANCEL

#DEFINE MB_ICON_QUESTION         0   // Message Box with icon QUESTION
#DEFINE MB_ICON_EXCLAMATION      1   // Message Box with icon EXCLAMATION
#DEFINE MB_ICON_STOP             2   // Message Box with icon STOP
#DEFINE MB_ICON_INFO             3   // Message Box with icon INFO

//                       ------------- Standard stats graph definitions -------
#DEFINE STYLE_BAR               1    //  Bars
#DEFINE STYLE_LINE              2    //  Lines
#DEFINE STYLE_BAR_3D            3    //  3D Bars
#DEFINE STYLE_LINE_3D           4    //  3D Lines



/***
*  BUTTONS
*/
#command @ <nLeft>,<nTop>,[<nRight>],[<nBottom>] BUTTON  ;
         [STYLE <nStyle>]                                ;
         [ACTION <bAction>]                              ;
         [ACCELERATOR <xAccelKey>]                       ;
         [DISPLAYUP <bDisplayUp>]                        ;
         [DISPLAYDN <bDisplayDn>]                        ;
         [CARGO <xCargoVal>]                             ;
         ATTACH <aButCtrls>                              ;
      => aButtonAdd(<nLeft>                             ,;
                    <nTop>                              ,;
                    <nRight>                            ,;
                    <nBottom>                           ,;
                    <nStyle>                            ,;
                    <bAction>                           ,;
                    <xAccelKey>                         ,;
                    <bDisplayUp>                        ,;
                    <bDisplayDn>                        ,;
                    <aButCtrls>                         ,;
                    <xCargoVal>                          ;
                   )


#command SHOW ALL BUTTONS <aCtrlButtons>                ;
       => ButtonShowAll(<aCtrlButtons>)


/***
*  SCROLLBARS
*/
#command @ <nLeft>,<nTop>,<nRight>,<nBottom> SCROLLBAR  ;
         [STYLE <nStyle>]                               ;
         [ACTIONUP <bActionUp>]                         ;
         [ACTIONDN <bActionDn>]                         ;
         [ACTIONBAR <bActionBar>]                       ;
         HANDLE <aScrollHandle>                         ;
         ATTACH <aButCtrls>                             ;
      => aScrollAdd(<nLeft>                            ,;
                    <nTop>                             ,;
                    <nRight>                           ,;
                    <nBottom>                          ,;
                    <nStyle>                           ,;
                    <bActionUp>                        ,;
                    <bActionDn>                        ,;
                    <bActionBar>                       ,;
                    <aButCtrls>                        ,;
                    <aScrollHandle>                     ;
                   )

/***
*  POTENTIOMETERS
*/
#command @ <nLeft>,<nTop> WIDTH <nWidth> POTENTIOMETER  ;
         SETGET <bValSetGet>                            ;
         [EXECUTE <bExecOnChange>]                      ;
         [RANGE <nValMin>,<nValMax>]                    ;
         [PICTURE <cValPic>]                            ;
         HANDLE <aPotentHandle>                         ;
         ATTACH <aButCtrls>                             ;
      => aPotAdd(<nLeft>                               ,;
                 <nTop>                                ,;
                 <nWidth>                              ,;
                 <bValSetGet>                          ,;
                 <bExecOnChange>                       ,;
                 <nValMin>,<nValMax>                   ,;
                 <cValPic>                             ,;
                 <aButCtrls>                           ,;
                 <aPotentHandle>                        ;
                )

/***
*  EXTENDED GETS
*/
                                     // ChkBox gets

#COMMAND                                  ;
   @ <row>,<col> SAY <SayExpr>            ;
     GET <var> CHECKBOX                   ;
     ATTACH <aButCtrls>                   ;
 =>                                       ;
   SETPOS(<row>,<col>) ;                  ;
   ChkAddGet( { |x| IF(x==NIL,            ;
                       <var>,             ;
                       <var> := x         ;
                      )                   ;
               } ,                        ;
               <(var)>,                   ;
               GetList,                   ;
               <aButCtrls>,               ;
               <SayExpr>                  ;
                )

                                     // Incremental gets

#COMMAND                                             ;
   @ <Row>,<Col> GET <Var>                           ;
   [<clauses,...>] INCREMENTAL ATTACH <aButCtrls>    ;
   [<moreclause,...>]                                ;
 =>                                                  ;
  @ <Row>,<Col> GET <Var>                            ;
  [<clauses>]                                        ;
  SEND reader:={|oGet , GetList , aCtrlButtons|      ;
       IncrGetReader(oGet , GetList , aCtrlButtons)} ;
  [<moreclause>]                                     ;
  ;IncreButton(ATAIL(GetList),<aButCtrls>,GetList)


                                     // Assist gets
#COMMAND                                                                 ;
   @ <Row>,<Col> GET <Var>                                               ;
   [<clauses,...>] ASSIST <bAssist> ATTACH <aButCtrls>                   ;
   [<moreclause,...>]                                                    ;
 =>                                                                      ;
  @ <Row>,<Col> GET <Var>                                                ;
  [<clauses>]                                                            ;
  SEND reader:={|oGet , GetList , aCtrlButtons|                          ;
       DropGetReader(oGet , GetList , aCtrlButtons,NIL,<bAssist>)}       ;
  [<moreclause>]                                                         ;
  ;DropButton(ATAIL(GetList),<aButCtrls>,GetList,{ |nX,nY,aBut| xDropSpecial(nX,nY,aBut,NIL,<bAssist>) } )


                                     // Drop box gets
#COMMAND                                                                    ;
   @ <Row>,<Col> GET <Var>                                                  ;
   [<clauses,...>] DROPBOX <dropValues> ATTACH <aButCtrls>                  ;
   [<moreclause,...>]                                                       ;
 =>                                                                         ;
  @ <Row>,<Col> GET <Var>                                                   ;
  [<clauses>]                                                               ;
  SEND reader:={|oGet , GetList , aCtrlButtons|                             ;
       DropGetReader(oGet , GetList , aCtrlButtons,<dropValues>,NIL)}       ;
  [<moreclause>]                                                            ;
  ;DropButton(ATAIL(GetList),<aButCtrls>,GetList,                           ;
              { |nX,nY,aButton| xGetDropBox(nX,nY,aButton,<dropValues>) })



                                     // Radios gets

#COMMAND                                  ;
   @ <row>,<col> GET <var>                ;
     RADIO <radio,...>                    ;
     [<horizontal:HORIZONTAL>]            ;
     [<allowempty:ALLOW EMPTY>]           ;
     ATTACH <aButCtrls>                   ;
 =>                                       ;
   SETPOS(<row>,<col>) ;                  ;
   RadioAddGets( { |x| IF(x==NIL,         ;
                          <var>,          ;
                          <var> := x      ;
                         )                ;
                  } ,                     ;
                  <(var)>,{<radio>},      ;
                  GetList,                ;
                  <aButCtrls>,            ;
                  <.horizontal.>,         ;
                  <.allowempty.>          ;
                ) ;                       ;
   RadioDisplay(ATAIL(GetList),GetList)

/***
*  EXTENDED READS FOR LIGHT LIB GRAPHICS
*/

#command READ SAVE [INITGET <nInitGet>] [BUTTONS <aCtrlButtons>]       ;
       => ReadModal(GetList, <nInitGet>, <aCtrlButtons>)

#command READ [INITGET <nInitGet>] [BUTTONS <aCtrlButtons>]            ;
      => ReadModal(GetList, <nInitGet>, <aCtrlButtons>)                ;
       ; GetList := {}
