#INCLUDE "llibi.ch"          // Include definitions file for Light Lib Images
#INCLUDE "llibitoo.ch"       // Include definitions file for Light Lib Images tools
#INCLUDE "llibg.ch"          // Include definitions file for Light Lib Graphics
#INCLUDE "llibgtoo.ch"       // Include definitions file for Light Lib Graphics tools

#INCLUDE "Inkey.ch"

STATIC aBmpBase      := {}   // Static array to load majors .BMP in an INIT function
STATIC lPalExclusive := .T.  // Palette is exclusive to one image

MEMVAR GetList               // Allow the use of public GetList variable
                             // without warnings

*
                             // Group Buttons
                             // Note : Here are some CA-CLIPPER
                             // to load major BMP
/***
*
*   LoadBmpImg()          Load the majors BMP when EXE start to avoid reloading
*
*
*/
*
#DEFINE EXCL_OF_DW                       1    // Check Get Empty- Pos Dw
#DEFINE EXCL_OF_UP                       2    // Check Get Full - Pos Up
#DEFINE EXCL_ON_DW                       3    // Check Get Empty- Pos Dw
#DEFINE EXCL_ON_UP                       4    // Check Get Full - Pos Up
#DEFINE CLOSWIN_UP                       5    // Close Window   - Pos Up
#DEFINE CLOSWIN_DW                       6    // Close Window   - Pos Dw


*
INIT FUNCTION LoadBmpImg()
                             // Load the majors BMP when EXE start to avoid reloading

   aBmpBase := {}

                             // Exclusive palette OFF
   AADD(aBmpBase,gBmpLoad( "CHECK_E.BMU" ))
   AADD(aBmpBase,gBmpLoad( "CHECK_E.BMD" ))

                             // Exclusive palette ON
   AADD(aBmpBase,gBmpLoad( "CHECK_F.BMU" ))
   AADD(aBmpBase,gBmpLoad( "CHECK_F.BMD" ))

                             // Close window
   AADD(aBmpBase,gBmpLoad( "CLOSEWIN.BMU"))
   AADD(aBmpBase,gBmpLoad( "CLOSEWIN.BMD"))


   RETURN (NIL)

*
                             // Image browses
                             // Note : Here are some CA-CLIPPER
                             // functions to allow browse of images
                             // scanning and printing
/***
*
*   ImgApplyKey()         Apply a key stroke to the iage browse
*   ImageVertBar()        Manage image vertical scroll bar
*   ImageHoriBar()        Manage image horizontal scroll bar
*   ImgDynaDisplay()      Manage current cursor position in image and display
*   ImgZoom()             Manage zoom operations
*   ImgRotate()           Manage rotation operations
*   lImgGauge()           Follow Light Lib Images operations
*   lLoadImgFile()        Load next/previous Image File
*   ScanImg()             Manage scanning operations
*   PrintImg()            Manage printing operations
*   aImgSelArea()         Selected an area with the mouse
*
*/

*
FUNCTION ImgApplyKey( oBrowseImg ,;  // Browse image pseudo object
                      nKey        ;  // Key value to be applied
                    )

                             // Apply a keystroke to the image browse

   DO CASE
                             // Down 1/10 of display window
      CASE nKey == K_DOWN
           oBrowseImg:nCurrentY += INT(oBrowseImg:nWinDispHeight/10)
           oBrowseImg:lIsStable := .F.

                             // Down one display window
      CASE nKey == K_PGDN
           oBrowseImg:nCurrentY += oBrowseImg:nWinDispHeight
           oBrowseImg:lIsStable := .F.

                             // Down to a maximum position
      CASE nKey == K_CTRL_PGDN
           oBrowseImg:nCurrentY := 10000
           oBrowseImg:lIsStable := .F.

                             // Up 1/10 of display window
      CASE nKey == K_UP
           oBrowseImg:nCurrentY -= INT(oBrowseImg:nWinDispHeight/10)
           oBrowseImg:lIsStable := .F.

                             // Up one display window
      CASE nKey == K_PGUP
           oBrowseImg:nCurrentY -= oBrowseImg:nWinDispHeight
           oBrowseImg:lIsStable := .F.

                             // Up to a maximum position
      CASE nKey == K_CTRL_PGUP
           oBrowseImg:nCurrentY := 0
           oBrowseImg:lIsStable := .F.

                             // Right 1/10 of display window
      CASE nKey == K_RIGHT
           oBrowseImg:nCurrentX += INT(oBrowseImg:nWinDispWidth/10)
           oBrowseImg:lIsStable := .F.

                             // Left 1/10 of display window
      CASE nKey == K_LEFT
           oBrowseImg:nCurrentX -= INT(oBrowseImg:nWinDispWidth/10)
           oBrowseImg:lIsStable := .F.

                             // Go to 0,0
      CASE nKey == K_HOME .OR. nKey == K_CTRL_HOME
           oBrowseImg:nCurrentX := 0
           oBrowseImg:nCurrentY := 0
           oBrowseImg:lIsStable := .F.

                             // Go to maximum position
      CASE nKey == K_END .OR. nKey == K_CTRL_END
           oBrowseImg:nCurrentX := 10000
           oBrowseImg:nCurrentY := 10000
           oBrowseImg:lIsStable := .F.

                             // Right one display window
      CASE nKey == K_CTRL_RIGHT
           oBrowseImg:nCurrentX += oBrowseImg:nWinDispWidth
           oBrowseImg:lIsStable := .F.

                             // Left one display window
      CASE nKey == K_CTRL_LEFT
           oBrowseImg:nCurrentX -= oBrowseImg:nWinDispWidth
           oBrowseImg:lIsStable := .F.

                             // Zoom options
      CASE CHR(nKey) $ '/*!-+'
           ImgZoom(nKey,oBrowseImg)
           oBrowseImg:lIsStable := .F.

                             // Rotate options
      CASE CHR(nKey) $ '><&'
           ImgRotate(nKey,oBrowseImg)
           oBrowseImg:lIsStable := .F.

      OTHERWISE


   ENDCASE

   RETURN (NIL)

*
FUNCTION ImageVertBar(nSens     ,;  // NIL, -1, 0, 1
                      nPercent  ,;  // NIL or 0 to 1
                      oBrowseImg ;  // Browse image pseudo object
                     )
                             // Manage image vertical scroll bar


   IF nSens <> NIL           // Move the pointer

      IF nSens == 0          // Set the position depending on %

         oBrowseImg:nCurrentY := INT(MAX(1,nPercent * (oBrowseImg:aDisPtr[LLI_IMG_HEIGHT]-oBrowseImg:nWinDispHeight)))
         oBrowseImg:lIsStable := .F.

      ELSEIF nSens == -1     // Move up

         oBrowseImg:nCurrentY -= INT(oBrowseImg:nWinDispHeight/10)
         oBrowseImg:lIsStable := .F.

      ELSEIF nSens == 1      // Move down

         oBrowseImg:nCurrentY += INT(oBrowseImg:nWinDispHeight/10)
         oBrowseImg:lIsStable := .F.

      ENDIF

   ELSE                      // nSens==NIL just mean : what is location in %

   ENDIF


                             // Current Y position must fit with image and
                             // window disply sizes
   IF oBrowseImg:nCurrentY  <= 0 .OR. oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] <= oBrowseImg:nWinDispHeight
      oBrowseImg:nCurrentY := 0
   ELSEIF oBrowseImg:nCurrentY >= oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] - oBrowseImg:nWinDispHeight
      oBrowseImg:nCurrentY := oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] - oBrowseImg:nWinDispHeight
   ENDIF


                             // Return location in %
   RETURN (oBrowseImg:nCurrentY/(oBrowseImg:aDisPtr[LLI_IMG_HEIGHT]-oBrowseImg:nWinDispHeight))



*
FUNCTION ImageHoriBar(nSens      ,;  // NIL, -1, 0, 1
                      nPercent   ,;  // NIL or 0 to 1
                      oBrowseImg  ;  // Browse image pseudo object
                     )
                             // Manage image horizontal scroll bar

   IF nSens <> NIL

      IF nSens == 0          // Move the pointer

                             // Set the position depending on %
         oBrowseImg:nCurrentX := INT(MAX(1,nPercent * (oBrowseImg:aDisPtr[LLI_IMG_WIDTH]-oBrowseImg:nWinDispWidth)))
         oBrowseImg:lIsStable := .F.


      ELSEIF nSens==1        // Move right

         oBrowseImg:nCurrentX += INT(oBrowseImg:nWinDispWidth/10)
         oBrowseImg:lIsStable := .F.

      ELSEIF nSens==-1       // Move left

         oBrowseImg:nCurrentX -= INT(oBrowseImg:nWinDispWidth/10)
         oBrowseImg:lIsStable := .F.

      ENDIF

   ELSE                      // nSens==NIL mean : what is location in %

   ENDIF

                             // Current X position must fit with image and
                             // window disply sizes
   IF oBrowseImg:nCurrentX <= 0 .OR. oBrowseImg:aDisPtr[LLI_IMG_WIDTH] <= oBrowseImg:nWinDispWidth
      oBrowseImg:nCurrentX := 0
   ELSEIF oBrowseImg:nCurrentX >= oBrowseImg:aDisPtr[LLI_IMG_WIDTH] - oBrowseImg:nWinDispWidth
      oBrowseImg:nCurrentX := oBrowseImg:aDisPtr[LLI_IMG_WIDTH] - oBrowseImg:nWinDispWidth
   ENDIF

                             // Return location in %
   RETURN (oBrowseImg:nCurrentX/(oBrowseImg:aDisPtr[LLI_IMG_WIDTH]-oBrowseImg:nWinDispWidth))


*
FUNCTION ImgDynaDisplay(oBrowseImg  ;  // Browse image pseudo object
                       )

                             // Manage current cursor position in image and
                             // display image if no action pending


                             // Current X position must fit with image and
                             // window disply sizes
   IF oBrowseImg:nCurrentX <= 0 .OR. oBrowseImg:aDisPtr[LLI_IMG_WIDTH] <= oBrowseImg:nWinDispWidth
      oBrowseImg:nCurrentX := 0
   ELSEIF oBrowseImg:nCurrentX >= oBrowseImg:aDisPtr[LLI_IMG_WIDTH] - oBrowseImg:nWinDispWidth
      oBrowseImg:nCurrentX := oBrowseImg:aDisPtr[LLI_IMG_WIDTH] - oBrowseImg:nWinDispWidth
   ENDIF

                             // Current Y position must fit with image and
                             // window disply sizes
   IF oBrowseImg:nCurrentY  <= 0 .OR. oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] <= oBrowseImg:nWinDispHeight
      oBrowseImg:nCurrentY := 0
   ELSEIF oBrowseImg:nCurrentY >= oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] - oBrowseImg:nWinDispHeight
      oBrowseImg:nCurrentY := oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] - oBrowseImg:nWinDispHeight
   ENDIF

                             // If no keyboard or action pending,
                             // execute image display
   IF EMPTY(NEXTKEY())


      mHide()                // Hide mouse before display image

      IF oBrowseImg:aDisPtr[LLI_IMG_WIDTH] <= oBrowseImg:nWinDispWidth .OR. oBrowseImg:aDisPtr[LLI_IMG_HEIGHT] <= oBrowseImg:nWinDispHeight
                             // If image is smaller than window, clear window
                             // before image display

         gRect(oBrowseImg:nWinDispX1 ,;
               oBrowseImg:nWinDispY1 ,;
               oBrowseImg:nWinDispX2 ,;
               oBrowseImg:nWinDispY2 ,;
               LLG_FILL              ,;
               7                     ,;
               LLG_MODE_SET           ;
              )

      ENDIF

                             // Put image on screen
      iPut( oBrowseImg:aDisPtr                                           ,;
            oBrowseImg:nCurrentX                                         ,;
            oBrowseImg:nCurrentY                                         ,;
            oBrowseImg:nCurrentX+oBrowseImg:nWinDispWidth-1              ,;
            oBrowseImg:nCurrentY+oBrowseImg:nWinDispHeight-1             ,;
            LLI_SCREEN, NIL                                              ,;
            oBrowseImg:nWinDispX1                                        ,;
            oBrowseImg:nWinDispY1                                        ,;
            {IF(lPalExclusive,LLI_PALETTE_EXCLUSIVE,LLI_PALETTE_SHARED)}  ;
          )

                             // Image is now stable
      oBrowseImg:lIsStable := .T.

      mShow()                // Show mouse

   ENDIF


   RETURN (NIL)



*
FUNCTION ImgZoom(nKey       ,;  // Key Value
                 oBrowseImg  ;  // Browse image pseudo object
                )
                             // Manage zoom operations
                             // We manage the original image oBrowseImg:aOriPtr
                             // and the diplayed image oBrowseImg:aDisPtr

   LOCAL aSelected      := {}// Selected coordinates
   LOCAL aSetSave       := {}


   IF nKey==ASC('-')
                             // Zoom backward 25%
      oBrowseImg:nScaleX   *= 1/1.25
      oBrowseImg:nScaleY   *= 1/1.25
      oBrowseImg:nCurrentX *= 1/1.25
      oBrowseImg:nCurrentY *= 1/1.25

   ELSEIF nKey==ASC('+')
                             // Zoom forward 25%
      oBrowseImg:nScaleX   *= 1.25
      oBrowseImg:nScaleY   *= 1.25
      oBrowseImg:nCurrentX *= 1.25
      oBrowseImg:nCurrentY *= 1.25

   ELSEIF nKey==ASC('*')
                             // Reset zoom factor's to 1
      oBrowseImg:nScaleX   := 1
      oBrowseImg:nScaleY   := 1
      oBrowseImg:nCurrentX := 0
      oBrowseImg:nCurrentY := 0

   ELSEIF nKey==ASC('/')
                             // Reset zoom factor's to fit image in window

      oBrowseImg:nScaleX   := oBrowseImg:nWinDispWidth  / oBrowseImg:aOriPtr[LLI_IMG_WIDTH]
      oBrowseImg:nScaleY   := oBrowseImg:nWinDispHeight / oBrowseImg:aOriPtr[LLI_IMG_HEIGHT]
      oBrowseImg:nCurrentX := 0
      oBrowseImg:nCurrentY := 0

   ELSEIF nKey==ASC('!')
                             // Reset zoom factor's to fit image in window width

      oBrowseImg:nScaleX   := oBrowseImg:nWinDispWidth  / oBrowseImg:aOriPtr[LLI_IMG_WIDTH]
      oBrowseImg:nScaleY   := oBrowseImg:nWinDispWidth  / oBrowseImg:aOriPtr[LLI_IMG_WIDTH]
      oBrowseImg:nCurrentX := 0
      oBrowseImg:nCurrentY := 0


   ELSEIF nKey==ASC('')

                             // Select area, starting with mouse at the middle
                             // of the image window
      aSelected := aImgSelArea( oBrowseImg:nWinDispX1                                                               ,;
                                oBrowseImg:nWinDispY1                                                               ,;
                                MIN(oBrowseImg:nWinDispX2,oBrowseImg:aDisPtr[LLI_IMG_WIDTH] +oBrowseImg:nWinDispX1) ,;
                                MIN(oBrowseImg:nWinDispY2,oBrowseImg:aDisPtr[LLI_IMG_HEIGHT]+oBrowseImg:nWinDispY1)  ;
                              )


                             // If area have been selected
      IF aSelected[AREA_SELECTED]

         aSelected[AREA_X1] := INT((aSelected[AREA_X1]-oBrowseImg:nWinDispX1)/oBrowseImg:nScaleX)
         aSelected[AREA_Y1] := INT((aSelected[AREA_Y1]-oBrowseImg:nWinDispY1)/oBrowseImg:nScaleY)
         aSelected[AREA_X2] := INT((aSelected[AREA_X2]-oBrowseImg:nWinDispX1)/oBrowseImg:nScaleX)
         aSelected[AREA_Y2] := INT((aSelected[AREA_Y2]-oBrowseImg:nWinDispY1)/oBrowseImg:nScaleY)


         oBrowseImg:nScaleX   := oBrowseImg:nWinDispWidth  / (aSelected[AREA_X2] - aSelected[AREA_X1])
         oBrowseImg:nScaleY   := oBrowseImg:nWinDispHeight / (aSelected[AREA_Y2] - aSelected[AREA_Y1])

         oBrowseImg:nCurrentX := (oBrowseImg:nCurrentX + aSelected[AREA_X1]) * oBrowseImg:nScaleX
         oBrowseImg:nCurrentY := (oBrowseImg:nCurrentY + aSelected[AREA_Y1]) * oBrowseImg:nScaleY

      ENDIF

   ENDIF

                             // Erase VMM item
   oBrowseImg:aDisPtr := iErase(oBrowseImg:aDisPtr)

                             // Set UDF call to a larger number because zoom
                             // operations are generaly longer than display
                             // operations
   aSetSave := iSet( LLI_SET_UDF , oBrowseImg:aUdfLong )

                             // Execute zoom to obtain oBrowseImg:aDisPtr from oBrowseImg:aOriPtr
   oBrowseImg:aDisPtr := iCopy( oBrowseImg:aOriPtr                                           ,;
                                0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE                           ,;
                                LLI_COPY_ZOOM                                                ,;
                                { INT(oBrowseImg:aOriPtr[LLI_IMG_WIDTH] *oBrowseImg:nScaleX) ,;
                                  INT(oBrowseImg:aOriPtr[LLI_IMG_HEIGHT]*oBrowseImg:nScaleY)  ;
                                }                                                             ;
                              )

                             // Reset UDF to standard value
   iSet( LLI_SET_UDF , aSetSave )

                             // Image is not stable
   oBrowseImg:lIsStable := .F.

   RETURN (NIL)


*
FUNCTION ImgRotate(nAngle     ,;  // Rotation angle
                   oBrowseImg  ;  // Browse image pseudo object
                  )
                             // Manage rotation operations
                             // We manage the original image oBrowseImg:aOriPtr
                             // and the diplayed image oBrowseImg:aDisPtr

   LOCAL nI       := 0
   LOCAL aSetSave := {}

                             // If we rotate X and Y axis, we need to swap scales values
   IF ABS(nAngle) == 90
      nI                 := oBrowseImg:nScaleX
      oBrowseImg:nScaleX := oBrowseImg:nScaleY
      oBrowseImg:nScaleY := nI
   ENDIF
                             // Set UDF call to a larger number because rotation
                             // operations are generaly longer than display
                             // operations
   aSetSave := iSet( LLI_SET_UDF , oBrowseImg:aUdfLong )

                             // Erase display copy of the image
   oBrowseImg:aDisPtr := iErase(oBrowseImg:aDisPtr)
                             // Point to Original with display Ptr
   oBrowseImg:aDisPtr := oBrowseImg:aOriPtr


   IF nAngle == 90
                             // Turn 90
      oBrowseImg:aOriPtr := iCopy( oBrowseImg:aDisPtr                   ,;
                                   0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE   ,;
                                   LLI_COPY_TURN                        ,;
                                   { LLI_TURN_90 }                       ;
                                 )


   ELSEIF nAngle == 180
                             // Turn 180
      oBrowseImg:aOriPtr := iCopy( oBrowseImg:aDisPtr                  ,;
                                   0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE  ,;
                                   LLI_COPY_TURN                       ,;
                                   { LLI_TURN_180 }                     ;
                                 )


   ELSEIF nAngle == -90
                             // Turn 270
      oBrowseImg:aOriPtr := iCopy( oBrowseImg:aDisPtr                  ,;
                                   0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE  ,;
                                   LLI_COPY_TURN                       ,;
                                   { LLI_TURN_270 }                     ;
                                 )
   ENDIF


   IF EMPTY(oBrowseImg:aOriPtr)

                             // Operation have been aborted during cycle
      oBrowseImg:aOriPtr := oBrowseImg:aDisPtr

                             // Duplicate image to allow image to be scaled
      oBrowseImg:aDisPtr := iCopy( oBrowseImg:aOriPtr, 0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, LLI_COPY_CLONE )

   ENDIF

                             // Reset UDF to standard value
   iSet( LLI_SET_UDF , aSetSave )

                             // Resize image to correct zoom
   ImgZoom(0,oBrowseImg)
   ImgDynaDisplay(oBrowseImg)


   RETURN (NIL)




*
FUNCTION lImgGauge(nCallMode     ,;   // Call mose ie LLI_UDF_INIT | IDLE ...
                   nValue        ,;   // Value from Light Lib Images
                   lVisualGauge  ,;   // Gauge is visual ?
                   oBrowseImg     ;   // Browse image pseudo object
                  )

                             // Display a gauge (if visual) and follow images
                             // operations

                             // Notes on iSet(LLI_SET_UDF, { MyBlock, nMaxVal } )
                             // functions.

                             // When you execute an Iget()|iPut()|iCopy() function,
                             // MyBlock is called.
                             // MyBlock is alway called a first time with a
                             //         LLI_UDF_INIT.
                             // MyBlock is alway called at the end with a
                             //         LLI_UDF_EXIT.
                             // In the idle cycle, MyBlock will be called
                             // the number of time you define in nMaxVal parameters.
                             // In the idle cycle, MyBlock should be called with an
                             // LLI_UDF_ERROR parameter. The second parameter should
                             // precise what kind of error is involved.
                             // There is no retry on errors. After an LLI_UDF_ERROR,
                             // Light Lib Images will call MyBlock with LLI_UDF_EXIT
                             // and will abort execution. For iGet()|iCopy() functions,
                             // the result will be a NIL pointer. For iPut() function,
                             // the result already displayed on screen or printer will
                             // stay but if iPut() is involved in a disk operation the
                             // disk file will be deleted.

   STATIC aHandle   := {}    // Handle on gauge

                             // By default, continue operations
   LOCAL  lContinue := LLI_UDF_CONT


   IF nCallMode == LLI_UDF_INIT

      IF lVisualGauge        // If gauge is visible

                             // Create a new gauge and
                             // set the number of futur calls
         lGauge(GAUGE_INIT                                                       ,;
                aHandle                                                          ,;
                GAUGE_HORI                                                       ,;
                INT(oBrowseImg:nWinDispY2/FONT_SIZE_Y)+1                         ,;
                INT(oBrowseImg:nWinDispX1/FONT_SIZE_X)                           ,;
                INT((oBrowseImg:nWinDispX2+1-oBrowseImg:nWinDispX1)/FONT_SIZE_X) ,;
                { || nValue }                                                    ,;  // Max value
                { || nGaugeCurValue() }                                           ;
               )

      ENDIF

   ELSEIF nCallMode == LLI_UDF_IDLE

      IF lVisualGauge        // If gauge is visible

                             // Set current value
         nGaugeCurValue(nValue)
                             // Display Gauge
         lGauge(GAUGE_IDLE,aHandle)

         IF NEXTKEY()==K_ESC

            IF nMessageBox("Do you really want to stop operation" ,;
                           "Abort"                                ,;
                           MB_YES_NO                              ,;
                           MB_ICON_STOP                            ;
                          ) == 1

               lContinue := LLI_UDF_ABORT

            ELSE

               lContinue := LLI_UDF_CONT

            ENDIF

         ENDIF

      ELSE

         IF NEXTKEY()<>0
            lContinue := LLI_UDF_ABORT
         ENDIF

      ENDIF

   ELSEIF nCallMode == LLI_UDF_EXIT

      IF lVisualGauge        // If gauge is visible

                             // Close Gauge
         lGauge(GAUGE_EXIT,aHandle)

      ENDIF

   ELSEIF nCallMode == LLI_UDF_ERROR

      DO CASE


         CASE nValue == LLI_ERROR_MEMORY_BAD_PARAMETERS

               nMessageBox("Bad parameters in a memory operation" ,;
                           "Memory allocation"                    ,;
                           MB_OK                                  ,;
                           MB_ICON_STOP                            ;
                        )


         CASE nValue == LLI_ERROR_MEMORY_NO_MORE_SPACE

               nMessageBox("No more space in the VMM "  ,;
                           "Memory allocation"          ,;
                           MB_OK                        ,;
                           MB_ICON_STOP                  ;
                        )


         CASE nValue == LLI_ERROR_SCREEN_BAD_PARAMETERS

               nMessageBox("Bad parameters in a display operation" ,;
                           "Screen display"                        ,;
                           MB_OK                                   ,;
                           MB_ICON_STOP                             ;
                        )


         CASE nValue == LLI_ERROR_SCREEN_BAD_MODE

               nMessageBox("Unable to display images;Bad video mode"  ,;
                           "Screen display"                           ,;
                           MB_OK                                      ,;
                           MB_ICON_STOP                                ;
                        )


         CASE nValue == LLI_ERROR_DISK_BAD_PARAMETERS

               nMessageBox("Bad parameters in a disk operation" ,;
                           "Disk operation"                     ,;
                           MB_OK                                ,;
                           MB_ICON_STOP                          ;
                        )


         CASE nValue == LLI_ERROR_DISK_NOT_AVAILABLE

               nMessageBox("Requested disk is not available" ,;
                           "Disk operation"                  ,;
                           MB_OK                             ,;
                           MB_ICON_STOP                       ;
                        )


         CASE nValue == LLI_ERROR_DISK_FULL

               nMessageBox("Requested disk is full" ,;
                           "Disk operation"         ,;
                           MB_OK                    ,;
                           MB_ICON_STOP              ;
                        )


         CASE nValue == LLI_ERROR_DISK_WRITE_PROTECTED

               nMessageBox("Requested disk is write protected" ,;
                           "Disk operation"                    ,;
                           MB_OK                               ,;
                           MB_ICON_STOP                         ;
                        )


         CASE nValue == LLI_ERROR_DISK_READ_WRITE_ERROR

               nMessageBox("Read/write error on requeted disk" ,;
                           "Disk operation"                    ,;
                           MB_OK                               ,;
                           MB_ICON_STOP                         ;
                        )


         CASE nValue == LLI_ERROR_DISK_BAD_FILE_FORMAT

               nMessageBox("Bad file format read image file" ,;
                           "Disk operation"                  ,;
                           MB_OK                             ,;
                           MB_ICON_STOP                       ;
                        )


         CASE nValue == LLI_ERROR_PRINTER_BAD_PARAMETERS


               nMessageBox("Bad parameters in a print operation" ,;
                           "Print operation"                     ,;
                           MB_OK                                 ,;
                           MB_ICON_STOP                           ;
                        )


         CASE nValue == LLI_ERROR_PRINTER_NOT_ONLINE

               nMessageBox("Printer is not on line" ,;
                           "Print operation"        ,;
                           MB_OK                    ,;
                           MB_ICON_STOP              ;
                        )


         CASE nValue == LLI_ERROR_PRINTER_PAPER_OUT

               nMessageBox("Paper out on printer" ,;
                           "Print operation"      ,;
                           MB_OK                  ,;
                           MB_ICON_STOP            ;
                        )


         CASE nValue == LLI_ERROR_SCANNER_BAD_PARAMETERS

               nMessageBox("Bad parameters in a scan operation" ,;
                           "Scan operation"                     ,;
                           MB_OK                                ,;
                           MB_ICON_STOP                          ;
                        )


         CASE nValue == LLI_ERROR_SCANNER_NOT_ONLINE

               nMessageBox("Scanner is not on line" ,;
                           "Scan operation"         ,;
                           MB_OK                    ,;
                           MB_ICON_STOP              ;
                        )

      ENDCASE


   ENDIF


   RETURN (lContinue)

*
STATIC FUNCTION nGaugeCurValue(nValue)
                             // Set Get function to allow Gauge use

   STATIC nCurValue := 0

   IF nValue<>NIL

      nCurValue := nValue

   ENDIF

   RETURN (nCurValue)

*
FUNCTION PalExcShaDisp(oBrowseImg,lDisplayUp)

                             // Display palette mode selected
                             // exclusive | shared

   IF lPalExclusive
                             // If palette is exclusive

      gBmpDisp(IF(lDisplayUp,aBmpBase[EXCL_ON_UP],aBmpBase[EXCL_ON_DW]),;
               oBrowseImg:nWinDispX2 + 1                               ,;
               oBrowseImg:nWinDispY2 + 1                                ;
              )

   ELSE                      // If palette is shared

      gBmpDisp(IF(lDisplayUp,aBmpBase[EXCL_OF_UP],aBmpBase[EXCL_OF_DW]),;
               oBrowseImg:nWinDispX2 + 1                               ,;
               oBrowseImg:nWinDispY2 + 1                                ;
              )

   ENDIF


   RETURN (NIL)

*
FUNCTION PalExcShaSwitch(oBrowseImg)

                             // Switch from exclusive to shared palette
                             // mode and vice versa
   lPalExclusive := !lPalExclusive

                             // Image is no longer stable. Need to be displayed
   oBrowseImg:lIsStable := .F.

                             // Palette exclusive|shared redisplay state
   PalExcShaDisp(oBrowseImg,.T.)

                             // Redisplay image with new palette
   ImgDynaDisplay(oBrowseImg)

   RETURN (NIL)


*
FUNCTION lLoadImgFile(oBrowseImg ,;
                      xSensName   ;  // Sens or filename
                     )
                             // Load next/previous Image File of specified
                             // file


   LOCAL lLoadOk      := .F. // Load is succesfull
   LOCAL aSetSave     := {}  // Save old Light Lib Images iSet(UDF)
   LOCAL cCurrentFile := ''  // File to be loaded

   IF VALTYPE(xSensName)=='N'

                             // If it is a numeric value, assume it's a 1 or a -1
      oBrowseImg:nImgInUse += xSensName

                             // Maximize and minimize value in a circular way
      IF oBrowseImg:nImgInUse > LEN(oBrowseImg:aImgAvail)
         oBrowseImg:nImgInUse := 1
      ELSEIF oBrowseImg:nImgInUse == 0
         oBrowseImg:nImgInUse := LEN(oBrowseImg:aImgAvail)
      ENDIF
                             // Retrieve filename to be loaded
      cCurrentFile := oBrowseImg:aImgAvail[oBrowseImg:nImgInUse]

   ELSE
                             // Assume it's a filename
      cCurrentFile := xSensName

   ENDIF

                             // Reset scales and offsets values
   oBrowseImg:nCurrentX      := 0
   oBrowseImg:nCurrentY      := 0
   oBrowseImg:nScaleX        := 1
   oBrowseImg:nScaleY        := 1


                             // Set UDF call to a larger number because load
                             // operations are generaly longer than display
                             // operations
   aSetSave := iSet( LLI_SET_UDF , oBrowseImg:aUdfLong )

                             // Erase VMM items
   oBrowseImg:aOriPtr := iErase(oBrowseImg:aOriPtr)
   oBrowseImg:aDisPtr := iErase(oBrowseImg:aDisPtr)

                             // Depending on extension, use the correct file
                             // format to load the image file
   IF '.BMP'$cCurrentFile

                             // Read .BMP file
      oBrowseImg:aOriPtr := iGet( LLI_DISK, LLI_DISK_BMP,  0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, { cCurrentFile } )

   ELSEIF '.PCX'$cCurrentFile

                             // Read .PCX file
      oBrowseImg:aOriPtr := iGet( LLI_DISK, LLI_DISK_PCX,  0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, { cCurrentFile } )

   ELSEIF '.TIF'$cCurrentFile

                             // Read .TIF file
      oBrowseImg:aOriPtr := iGet( LLI_DISK, LLI_DISK_TIF,  0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, { cCurrentFile } )

   ENDIF

                             // An empty pointer means load failed for
                             // any reason (fileformat, no file ...)

                             // If the pointer is not empty
   IF !EMPTY(oBrowseImg:aOriPtr)

      lLoadOk :=.T.          // Load is successfull

                             // Image is not stable
      oBrowseImg:lIsStable := .F.

                             // Duplicate image to allow image to be scaled
      oBrowseImg:aDisPtr := iCopy( oBrowseImg:aOriPtr, 0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE, LLI_COPY_CLONE )

   ENDIF

                             // Reset UDF to standard value
   iSet( LLI_SET_UDF , aSetSave )

   RETURN (lLoadOk)


*
FUNCTION ScanImg(oBrowseImg  ;  // Browse image pseudo object
                )
                             // Scan an image and store it in a file

   LOCAL nTop        :=  5   // Window coordinates
   LOCAL nLeft       :=  9   //
   LOCAL nBottom     := 26   //
   LOCAL nRight      := 55   //

                             // Save environnement
   LOCAL xSave       := SAVESCREEN(nTop,nLeft,nBottom,nRight)
   LOCAL cSaveColor  := SETCOLOR("N/W,W+/B,,G,W+/N")

                             // Define some defaults values to be modified
   STATIC nScanX1    := 0
   STATIC nScanY1    := 0
   STATIC nScanX2    := 21.0
   STATIC nScanY2    := 29.7

   STATIC cScanType  := "HP Mono    "
   STATIC nScanLumi  := 0
   STATIC nScanCont  := 0

   STATIC cFileForm  := "TIF"
   STATIC lFileComp  := .T.
   STATIC cFileName  := 'MyFile  '

   STATIC cScanDens  := ' 75'

                             // Create an empty handler for buttons
   LOCAL aWinButtons := {}


                             // Display a 3D box
   DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Empty the GetList
   GetList := {}

                             // Get size informations
   @ nTop+02 , nLeft+04 SAY "From X     " GET nScanX1 PICTURE '99.9' INCREMENTAL ATTACH aWinButtons VALID (nScanX1>=0 .AND. nScanX1<nScanX2)
   @ nTop+03 , nLeft+04 SAY "     Y     " GET nScanY1 PICTURE '99.9' INCREMENTAL ATTACH aWinButtons VALID (nScanY1>=0 .AND. nScanY1<nScanX2)
   @ nTop+04 , nLeft+04 SAY "To   X     " GET nScanX2 PICTURE '99.9' INCREMENTAL ATTACH aWinButtons VALID (nScanX2>nScanX1 .AND. nScanX2<=21.0)
   @ nTop+05 , nLeft+04 SAY "     Y     " GET nScanY2 PICTURE '99.9' INCREMENTAL ATTACH aWinButtons VALID (nScanY2>nScanY1 .AND. nScanY2<=29.7)
   GetBox(nTop+02,nLeft+04,nTop+05,nLeft+21,4,'Size')


                             // Get scan type informations
   @ nTop+07 , nLeft+04 SAY "Mode" GET cScanType RADIO "HP Mono    ","HP Dither 4","HP Dither 8","HP Gray 256","HP Coul 256" ATTACH aWinButtons
   GetBox(nTop+07,nLeft+04,nTop+11,nLeft+21,4,'Scan mode')

                             // Get Density informations
   @ nTop+02 , nLeft+25 SAY "Density     " GET cScanDens RADIO " 75","100","150","300" ATTACH aWinButtons
   GetBox(nTop+02,nLeft+25,nTop+05,nLeft+42,4,'Scan Density')

                             // Get file informations
   @ nTop+07 , nLeft+25 SAY "FileName " GET cFileName PICTURE 'XXXXXXXX'
   @ nTop+08 , nLeft+33 SAY "Compress"  GET lFileComp CHECKBOX ATTACH aWinButtons
   @ nTop+09 , nLeft+25 SAY "Format "   GET cFileForm RADIO "TIF","PCX","BMP" ATTACH aWinButtons
   GetBox(nTop+07,nLeft+25,nTop+11,nLeft+42,4,'File')


                             // Add a potentiometer to manage luminosity
   @ nTop+13,nLeft+04 SAY "Black       "+CHR(27)+" Luminosity "+CHR(26)+"        White"
   @ nLeft+04,nTop+14 WIDTH 39                                   ;
     POTENTIOMETER                                               ;
     SETGET { |nVal| IF(nVal==NIL,nScanLumi,nScanLumi:=nVal) }   ;
     RANGE -100,+100                                             ;
     PICTURE '9999'                                              ;
     HANDLE {}                                                   ;
     ATTACH aWinButtons

                             // Add a potentiometer to manage contrast
   @ nTop+15,nLeft+04 SAY "Less        "+CHR(27)+"  Contrast  "+CHR(26)+"         More"
   @ nLeft+04,nTop+16 WIDTH 39                                   ;
     POTENTIOMETER                                               ;
     SETGET { |nVal| IF(nVal==NIL,nScanCont,nScanCont:=nVal) }   ;
     RANGE -100,+100                                             ;
     PICTURE '9999'                                              ;
     HANDLE {}                                                   ;
     ATTACH aWinButtons
   GetBox(nTop+13,nLeft+04,nTop+16,nLeft+42,4,'Luminosity')


                             // Add a "CANCEL" button to emulate K_ESC
   @ (nLeft+10) * FONT_SIZE_X                ,;
     (nBottom-03) * FONT_SIZE_Y             ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_KEY                    ;
     ACTION K_ESC                             ;
     ACCELERATOR K_ALT_L                      ;
     DISPLAYUP gBmpLoad("CANCEL.BMU")         ;
     DISPLAYDN gBmpLoad("CANCEL.BMD")         ;
     ATTACH aWinButtons

                             // Add a "SCAN" button to launch scanning
   @ (nLeft+30) * FONT_SIZE_X                ,;
     (nBottom-03) * FONT_SIZE_Y             ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_RELEASE                ;
     ACTION { || ExecScan(oBrowseImg, nScanX1, nScanY1, nScanX2, nScanY2, cScanType, nScanLumi, nScanCont, cFileForm, lFileComp, cFileName, cScanDens) } ;
     ACCELERATOR K_ALT_O                      ;
     DISPLAYUP gBmpLoad("SCAN.BMU")           ;
     DISPLAYDN gBmpLoad("SCAN.BMD")           ;
     ATTACH aWinButtons

                             // Add a close window button
   @ nLeft*FONT_SIZE_X                       ,;
     nTop *FONT_SIZE_Y                      ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_KEY                    ;
     ACTION K_ESC                             ;
     DISPLAYUP aBmpBase[CLOSWIN_UP]           ;
     DISPLAYDN aBmpBase[CLOSWIN_DW]           ;
     ATTACH aWinButtons

                             // Show all buttons in aWinButtons list
   SHOW ALL BUTTONS aWinButtons

                             // Read gets, starting by third get, managing
                             // aWinButtons buttons
   READ BUTTONS aWinButtons

                             // Restore window
   RESTSCREEN(nTop,nLeft,nBottom,nRight,xSave)
   SETCOLOR(cSaveColor)      // Restore colors

   RETURN (NIL)


*
STATIC FUNCTION ExecScan(oBrowseImg    ,;  // Browse image pseudo object
                         nScanX1       ,;  // Coordinates of the
                         nScanY1       ,;  // rectangle to be scanned
                         nScanX2       ,;  // in millimeters
                         nScanY2       ,;  //
                         cScanType     ,;  // Type of scanner to be used
                         nScanLumi     ,;  // Luminosity to be used
                         nScanCont     ,;  // Contrast to be used
                         cFileForm     ,;  // File format to be used
                         lFileComp     ,;  // Need to compress file ?
                         cFileName     ,;  // File name to be used
                         cScanDens      ;  // Scan density to be used
                        )


   LOCAL aPtr                // Image pointer to grab scanner informations

                             // Default values to be used
   LOCAL nScanType := LLI_SCANNER_HP_BW
   LOCAL nScanDens := LLI_DPI_75
   LOCAL nFileForm := LLI_DISK_TIF
   LOCAL nFileComp := LLI_DISK_COMPRESS_NIL
   LOCAL aSetSave     := {}

                             // Determine scan type
   IF cScanType == "HP Mono    "

      nScanType := LLI_SCANNER_HP_BW

   ELSEIF cScanType == "HP Dither 4"

      nScanType := LLI_SCANNER_HP_DITHERED_4

   ELSEIF cScanType == "HP Dither 8"

      nScanType := LLI_SCANNER_HP_DITHERED_8

   ELSEIF cScanType == "HP Gray 256"

      nScanType := LLI_SCANNER_HP_GRAY_256

   ELSEIF cScanType == "HP Coul 256"

      nScanType := LLI_SCANNER_HP_COLOR_256

   ENDIF

                             // Determine scan density
   IF cScanDens == " 75"

      nScanDens := LLI_DPI_75

   ELSEIF cScanDens == "100"

      nScanDens := LLI_DPI_100

   ELSEIF cScanDens == "150"

      nScanDens := LLI_DPI_150

   ELSEIF cScanDens == "300"

      nScanDens := LLI_DPI_300

   ENDIF

                             // Determine file format
   IF cFileForm == "TIF"

      cFileName := ALLTRIM(cFileName)+'.TIF'

      nFileForm := LLI_DISK_TIF

      IF lFileComp
         nFileComp := LLI_DISK_COMPRESS_RLE
      ELSE
         nFileComp := LLI_DISK_COMPRESS_NIL
      ENDIF


   ELSEIF cFileForm == "BMP"

      cFileName := ALLTRIM(cFileName)+'.BMP'

      nFileForm := LLI_DISK_BMP
                             // No compression available for BMP files
      nFileComp := LLI_DISK_COMPRESS_NIL


   ELSEIF cFileForm == "PCX"

      cFileName := ALLTRIM(cFileName)+'.PCX'

      nFileForm := LLI_DISK_PCX
                             // PCX files are alway RLE compressed
      nFileComp := LLI_DISK_COMPRESS_RLE

   ENDIF

                             // Set UDF call to a larger number because load
                             // operations are generaly longer than display
                             // operations
   aSetSave := iSet( LLI_SET_UDF , oBrowseImg:aUdfLong )

                             // Lauch the scan operation
   aPtr := iGet( LLI_SCANNER      ,;
                 nScanType        ,;
                 nScanX1*10       ,;
                 nScanY1*10       ,;
                 nScanX2*10       ,;
                 nScanY2*10       ,;
                 { nScanDens      ,;
                   nScanLumi      ,;
                   nScanCont       ;
                 }                 ;
               )

   IF !EMPTY(aPtr)           // IF the scan have been succesfull

                             // Save the image in filename with good fileformat
                             // and compression mode
      iPut( aPtr                                    ,;
            0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE      ,;
            LLI_DISK, nFileForm                     ,;
            0 , 0                                   ,;
            { cFileName  , nFileComp }               ;
         )

   ENDIF
                             // Erase image from VMM
   aPtr := iErase(aPtr)

                             // Reset UDF to previous value
   iSet( LLI_SET_UDF , aSetSave )

   RETURN (NIL)

*
FUNCTION PrintImg(oBrowseImg  ;  // Browse image pseudo object
                 )
                             // Print the current image

   LOCAL nTop        :=  5   // Window coordinates
   LOCAL nLeft       :=  9   //
   LOCAL nBottom     := 24   //
   LOCAL nRight      := 40   //

                             // Save settings
   LOCAL xSave       := SAVESCREEN(nTop,nLeft,nBottom,nRight)
   LOCAL cSaveColor  := SETCOLOR("N/W,W+/B,,G,W+/N")

                             // Define some values to be modified
   STATIC cPrintWhat  := 'All '
   STATIC cPrintWhere := '1'
   STATIC cPrintType  := 'HP  '
   STATIC cPrintDens  := 'Auto'
   STATIC cFileName   := 'MyFile  '


                             // Create an empty handler for buttons
   LOCAL aWinButtons := {}

                             // Display a 3D box
   DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Empty the GetList
   GetList := {}

                             // Get what to print informations
   @ nTop+02 , nLeft+04 SAY "What    " GET cPrintWhat RADIO 'All ','Area' HORIZONTAL ATTACH aWinButtons
   GetBox(nTop+02,nLeft+04,nTop+02,nLeft+27,4,'Print what')

                             // Get where to print informations
   @ nTop+04 , nLeft+04 SAY "Where   " GET cPrintWhere RADIO '1','2','3','F' HORIZONTAL ATTACH aWinButtons
   GetBox(nTop+04,nLeft+04,nTop+04,nLeft+27,4,'Print where')

                             // Get printer type informations
   @ nTop+06 , nLeft+04 SAY "Type    " GET cPrintType RADIO 'HP  ','POST' HORIZONTAL ATTACH aWinButtons
   GetBox(nTop+06,nLeft+04,nTop+06,nLeft+27,4,'Print type')

                             // Get Density informations
   @ nTop+08 , nLeft+04 SAY "Density " GET cPrintDens RADIO 'Auto','  75',' 100',' 150',' 300' ATTACH aWinButtons
   GetBox(nTop+08,nLeft+04,nTop+12,nLeft+27,4,'Print density')

                             // Get file informations
   @ nTop+14 , nLeft+04 SAY "FileName" GET cFileName PICTURE 'XXXXXXXX' WHEN cPrintWhere=="F"
   GetBox(nTop+14,nLeft+04,nTop+14,nLeft+27,4,'FileName')


                             // Add a "CANCEL" button to emulate K_ESC
   @ (nLeft+05) * FONT_SIZE_X                ,;
     (nBottom-03) * FONT_SIZE_Y             ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_KEY                    ;
     ACTION K_ESC                             ;
     ACCELERATOR K_ALT_L                      ;
     DISPLAYUP gBmpLoad("CANCEL.BMU")         ;
     DISPLAYDN gBmpLoad("CANCEL.BMD")         ;
     ATTACH aWinButtons

                             // Add a "PRINT" button to launch printing
   @ (nLeft+20) * FONT_SIZE_X                ,;
     (nBottom-03) * FONT_SIZE_Y             ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_RELEASE                ;
     ACTION { || ExecPrint(oBrowseImg, cPrintWhat, cPrintWhere, cPrintType, cPrintDens, cFileName ) } ;
     ACCELERATOR K_ALT_O                      ;
     DISPLAYUP gBmpLoad("PRINT.BMU")          ;
     DISPLAYDN gBmpLoad("PRINT.BMD")          ;
     ATTACH aWinButtons

                             // Add a close window button
   @ nLeft*FONT_SIZE_X                       ,;
     nTop *FONT_SIZE_Y                      ,,;
     BUTTON                                   ;
     STYLE BUTTON_TYPE_KEY                    ;
     ACTION K_ESC                             ;
     DISPLAYUP aBmpBase[CLOSWIN_UP]           ;
     DISPLAYDN aBmpBase[CLOSWIN_DW]           ;
     ATTACH aWinButtons

                             // Show all buttons in aWinButtons list
   SHOW ALL BUTTONS aWinButtons

                             // Read gets, starting by third get, managing
                             // aWinButtons buttons
   READ BUTTONS aWinButtons

                             // Restore window
   RESTSCREEN(nTop,nLeft,nBottom,nRight,xSave)
   SETCOLOR(cSaveColor)      // Restore colors

   RETURN (NIL)


*
STATIC FUNCTION ExecPrint( oBrowseImg    ,; // Browse image pseudo object
                           cPrintWhat    ,; // Print all the file or a part
                           cPrintWhere   ,; // Print to port 1/2/3 or disk file
                           cPrintType    ,; // Type of printer to be used
                           cPrintDens    ,; // Printing density
                           cFileName      ; // File name if needed
                          )

                             // By default print all the image
   LOCAL nPrintX1    := LLI_FULL_SIZE
   LOCAL nPrintY1    := LLI_FULL_SIZE
   LOCAL nPrintX2    := LLI_FULL_SIZE
   LOCAL nPrintY2    := LLI_FULL_SIZE

                             // Partial selection
   LOCAL nLeft       := 0
   LOCAL nTop        := 0
   LOCAL nRight      := 0
   LOCAL nBottom     := 0
   LOCAL xSaveScreen := NIL
   LOCAL xSaveCursor := NIL
   LOCAL aTmpPtr     := NIL
   LOCAL aSelected   := {}   // Selected coordinates
   LOCAL cTitle      := ''

                             // defaults values
   LOCAL xPrintWhere := 1
   LOCAL nPrintType  := LLI_PRINTER_PCL4_BW
   LOCAL nPrintDens  := LLI_DPI_75

   LOCAL aSetSave     := {}

                             // Area to be printed must be choose here
   IF cPrintWhat == 'Area'

      nTop    := 07
      nLeft   := 13
      nBottom := 22
      nRight  := nLeft + 2 * INT((nBottom-nTop+1)/oBrowseImg:aOriPtr[LLI_IMG_HEIGHT]*oBrowseImg:aOriPtr[LLI_IMG_WIDTH])

                             // Save environnement
      xSaveCursor := SETCURSOR(0)
      xSaveScreen := SAVESCREEN(nTop,nLeft,nBottom,nRight)

                             // Display a 3D box
      DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Display a title in box frame
      cTitle := LEFT("Use mouse to select area",nRight-nLeft+1)
      gWriteAt( ( nLeft+INT(((nRight-nLeft+1)-LEN(cTitle))/2 ))* FONT_SIZE_X ,;
                nTop * FONT_SIZE_Y                                           ,;
                cTitle                                                       ,;
                1                                                            ,;
                LLG_MODE_SET                                                  ;
              )


                             // Execute zoom to fit original image
                             // in a small select window
      aTmpPtr := iCopy( oBrowseImg:aOriPtr                                           ,;
                        0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE                           ,;
                        LLI_COPY_ZOOM                                                ,;
                        { INT((nRight-nLeft-3)*FONT_SIZE_X)                          ,;
                          INT((nBottom-nTop-1)*FONT_SIZE_Y)                           ;
                        }                                                             ;
                      )

                             // Put small image on screen
      iPut( aTmpPtr                                                      ,;
            0, 0, LLI_FULL_SIZE, LLI_FULL_SIZE                           ,;
            LLI_SCREEN, NIL                                              ,;
            (nLeft + 2)*FONT_SIZE_X                                      ,;
            (nTop  + 1)*FONT_SIZE_Y                                      ,;
            {IF(lPalExclusive,LLI_PALETTE_EXCLUSIVE,LLI_PALETTE_SHARED)}  ;
          )

                             // Select area, starting with mouse at the middle
                             // of the image window
      aSelected := aImgSelArea( (nLeft   + 2)*FONT_SIZE_X ,;
                                (nTop    + 1)*FONT_SIZE_Y ,;
                                (nRight  - 1)*FONT_SIZE_X ,;
                                (nBottom    )*FONT_SIZE_Y  ;
                              )


      IF aSelected[AREA_SELECTED]
                             // If area have been selected

         nPrintX1 := INT((aSelected[AREA_X1]-(nLeft+2)*FONT_SIZE_X)/(INT((nRight-nLeft-3)*FONT_SIZE_X)/oBrowseImg:aOriPtr[LLI_IMG_WIDTH]))
         nPrintY1 := INT((aSelected[AREA_Y1]-(nTop +1)*FONT_SIZE_Y)/(INT((nBottom-nTop-1)*FONT_SIZE_Y)/oBrowseImg:aOriPtr[LLI_IMG_HEIGHT]))
         nPrintX2 := INT((aSelected[AREA_X2]-(nLeft+2)*FONT_SIZE_X)/(INT((nRight-nLeft-3)*FONT_SIZE_X)/oBrowseImg:aOriPtr[LLI_IMG_WIDTH]))
         nPrintY2 := INT((aSelected[AREA_Y2]-(nTop +1)*FONT_SIZE_Y)/(INT((nBottom-nTop-1)*FONT_SIZE_Y)/oBrowseImg:aOriPtr[LLI_IMG_HEIGHT]))

      ELSE
                              // If no aera selected, use full size
         nPrintX1 := LLI_FULL_SIZE
         nPrintY1 := LLI_FULL_SIZE
         nPrintX2 := LLI_FULL_SIZE
         nPrintY2 := LLI_FULL_SIZE

      ENDIF

                             // Erase temporary image
      aTmpPtr := iErase(aTmpPtr)

                             // Restore window
      RESTSCREEN(nTop,nLeft,nBottom,nRight,xSaveScreen)
      SETCURSOR(xSaveCursor)

   ENDIF

                             // Determine where to print
   IF cPrintWhere == "1"
                             // Parallel port 1
      xPrintWhere := LLI_PRINTER_LPT1

   ELSEIF cPrintWhere == "2"
                             // Parallel port 2
      xPrintWhere := LLI_PRINTER_LPT2

   ELSEIF cPrintWhere == "3"
                             // Parallel port 3
      xPrintWhere := LLI_PRINTER_LPT3

   ELSEIF cPrintWhere == "F"
                             // Print to file
      xPrintWhere := cFileName

   ENDIF

                             // Determine printer type
   IF cPrintType  == 'HP  '

      nPrintType := LLI_PRINTER_PCL4_BW

   ELSEIF cPrintType  == 'POST'

      nPrintType := LLI_PRINTER_POST_BW

   ENDIF


                             // Determine printing density
   IF cPrintDens == 'Auto'

      nPrintDens := oBrowseImg:aOriPtr[LLI_IMG_DPI]

   ELSEIF cPrintDens == '  75'

      nPrintDens := LLI_DPI_75

   ELSEIF cPrintDens == ' 100'

      nPrintDens := LLI_DPI_100

   ELSEIF cPrintDens == ' 150'

      nPrintDens := LLI_DPI_150

   ELSEIF cPrintDens == ' 300'

      nPrintDens := LLI_DPI_300

   ENDIF

                             // Set UDF call to a larger number because load
                             // operations are generaly longer than display
                             // operations
   aSetSave := iSet( LLI_SET_UDF , oBrowseImg:aUdfLong )

                             // Lauch the print operation
   iPut( oBrowseImg:aOriPtr  ,;
         nPrintX1            ,;
         nPrintY1            ,;
         nPrintX2            ,;
         nPrintY2            ,;
         LLI_PRINTER         ,;
         nPrintType          ,;
         0                   ,;
         0                   ,;
         { xPrintWhere       ,;
           nPrintDens        ,;
           LLI_PRINTER_EJECT  ;
         }                    ;
       )

                             // Reset UDF to standard value
   iSet( LLI_SET_UDF , aSetSave )

   RETURN (NIL)

*
FUNCTION aImgSelArea( nTopX ,;         // Max coordinates for rectangle
                      nTopY ,;         // to be selected
                      nBotX ,;         //
                      nBotY  ;         //
                    )
                             // Selected an area with the mouse

                             // Selection aborted ?
   LOCAL lSelected := .F.
                             // Mouse state array handler
   LOCAL aMouse     := mState()
   LOCAL aMouseClip := {}    // Mouse clipping save array
                             // Temporary swap value
   LOCAL nValue     := 0
                             // Mousable rectangle color
   LOCAL nColor     := 3
                             // Is mouse visible ?
   LOCAL lVisible   := aMouse[LLM_STATE_VISIBLE]

                             // Rectangle coordinates
   LOCAL nX1        := INT((nTopX+nBotX)/2)
   LOCAL nY1        := INT((nTopY+nBotY)/2)
   LOCAL nX2        := 0
   LOCAL nY2        := 0


                             // Wait for mouse to released if needed
   DO WHILE (aMouse[LLM_STATE_LEFT]+aMouse[LLM_STATE_RIGHT])==LLM_BUTTON_DOWN
      aMouse := mState()
   ENDDO

                             // Display the mouse in the midle
   mShow(LLM_COOR_GRAPH,nX1,nY1)

                             // Limitate mouse to allowed area and save
                             // previous clipping area
   aMouseClip := mSetClip( nTopX           ,;
                           nTopY           ,;
                           nBotX           ,;
                           nBotY           ,;
                           LLM_COOR_GRAPH                                                                       ;
                         )


                             // Wait for a first clic
   DO WHILE (aMouse[LLM_STATE_LEFT]+aMouse[LLM_STATE_RIGHT])==LLM_BUTTON_UP
      aMouse := mState()
   ENDDO
                             // If left button is down, --> drag
   IF aMouse[LLM_STATE_LEFT]==LLM_BUTTON_DOWN .AND. aMouse[LLM_STATE_RIGHT]==LLM_BUTTON_UP

                             // Initialize rectangle coordinates
      nX2 := nX1 := aMouse[LLM_STATE_X]
      nY2 := nY1 := aMouse[LLM_STATE_Y]


                             // hide the mouse
      mHide()

                             // Draw the select rectangle in XOR mode
      gRect(nX1,nY1,nX2,nY2,LLG_FRAME,nColor,LLG_MODE_XOR)

                             // Show the mouse
      mShow()

                             // Loop until mouse button is released
      DO WHILE aMouse[LLM_STATE_LEFT]==LLM_BUTTON_DOWN .AND. aMouse[LLM_STATE_RIGHT]==LLM_BUTTON_UP

         aMouse := mState()  // Collect mouse status

                             // if mouse have been moved, upgrade rectangle
         IF ((nX2 <> aMouse[LLM_STATE_X]) .OR. (nY2 <> aMouse[LLM_STATE_Y]))

                             // hide the mouse
            mHide()
                             // Redraw old rectangle in XOR mode to clear
            gRect(nX1,nY1,nX2,nY2,LLG_FRAME,nColor,LLG_MODE_XOR)

                             // Draw the new one in XOR mode
            gRect(nX1,nY1,aMouse[LLM_STATE_X],aMouse[LLM_STATE_Y],LLG_FRAME,nColor,LLG_MODE_XOR)

                             // Show the mouse
            mShow()
                             // Set new values
            nX2 := aMouse[LLM_STATE_X]
            nY2 := aMouse[LLM_STATE_Y]

         ENDIF

      ENDDO
                             // Left button have been released
                             // Hide the mouse
      mHide()
                             // Redraw the rectangle to hide it
      gRect(nX1,nY1,nX2,nY2,LLG_FRAME,nColor,LLG_MODE_XOR)

                             // Show the mouse
      mShow()
                             // Selection is done
      lSelected := .T.

   ENDIF


   IF nX1 > nX2              // Swap values if necessary
      nValue := nX2
      nX2    := nX1
      nX1    := nValue
   ENDIF

   IF nY1 > nY2              // Sawp values if necessary
      nValue := nY2
      nY2    := nY1
      nY1    := nValue
   ENDIF

                             // Restore previous mouse clipping area
   mSetClip(aMouseClip[1]   ,;
            aMouseClip[2]   ,;
            aMouseClip[3]   ,;
            aMouseClip[4]   ,;
            LLM_COOR_GRAPH   ;
           )

   IF lVisible               // IF mouse was visible at start show it
      mShow()
   ELSE                      // ELSE hide it
      mHide()
   ENDIF

   RETURN ({lSelected,nX1,nY1,nX2,nY2})


*
                             // Images viewer
                             // Note : Here are some CA-CLIPPER
                             // functions to allow view images on screen
/***
*
*   ImageView()           View an image on screen
*
*/

*
FUNCTION ImageViewer(nTop         ,; // Box coordinates used to display
                     nLeft        ,; // the image
                     nBottom      ,; //
                     nRight       ,; //
                     cImgFileName  ; // Image file name
                    )

   LOCAL nKey                // Keystroke
   LOCAL lMore := .T.        // Loop control

   LOCAL aWinButtons  := {}  //  Arrays to handle buttons
   LOCAL aVertScroll  := {}  //  and scrolls bars
   LOCAL aHoriScroll  := {}  //

                             // General savings
   LOCAL xScreen     := SAVESCREEN( nTop, nLeft, nBottom, nRight)
   LOCAL cColorSave  := SETCOLOR('N/W,W+/G,W,W,GR+/W')
   LOCAL nCursoSave  := SETCURSOR(0)

                             // Browse pseudo object
   LOCAL oBrowseImg  := ARRAY(BROWSE_IMG_SIZE)

   oBrowseImg:lIsStable      := .F.
   oBrowseImg:nCurrentX      := 0
   oBrowseImg:nCurrentY      := 0
   oBrowseImg:nScaleX        := 0
   oBrowseImg:nScaleY        := 0

                             // No image loaded, ptr are empty
   oBrowseImg:aOriPtr        := {}
   oBrowseImg:aDisPtr        := {}

                             // Compute display window pixels values
   oBrowseImg:nWinDispX1     := FONT_SIZE_X * ( nLeft + 2 )
   oBrowseImg:nWinDispY1     := FONT_SIZE_Y * ( nTop + 3 )
   oBrowseImg:nWinDispX2     := FONT_SIZE_X * ( nRight - 3 )  - 1
   oBrowseImg:nWinDispY2     := FONT_SIZE_Y * ( nBottom - 1 ) - 1
   oBrowseImg:nWinDispWidth  := oBrowseImg:nWinDispX2 - oBrowseImg:nWinDispX1 + 1
   oBrowseImg:nWinDispHeight := oBrowseImg:nWinDispY2 - oBrowseImg:nWinDispY1 + 1

                             // Fill the images to be browsed
   oBrowseImg:aImgAvail      := {}
   oBrowseImg:nImgInUse      := 0

                             // Set the block to standard apply key
   oBrowseImg:bApplyKey      := { |oBrowse,nKey| ImgApplyKey( oBrowse, nKey ) }
   oBrowseImg:aUdfShort      := { { | nStatus , nCall | lImgGauge( nStatus , nCall, .F., oBrowseImg ) } , 10 }
   oBrowseImg:aUdfLong       := { { | nStatus , nCall | lImgGauge( nStatus , nCall, .T., oBrowseImg ) } , 20 }

                             // -LLIBG- Display 3D box
   DISPBOX(nTop,nLeft,nBottom,nRight,LLG_BOX_GRAY_SQUARE)

                             // Set gauge function
   iSet( LLI_SET_UDF , oBrowseImg:aUdfShort )

                             // If loading next image is a succes
   IF lLoadImgFile( oBrowseImg , cImgFileName)

                             // -LLIBG- Add vertical scroll bar
      @ (nRight-3)*FONT_SIZE_X                                                ,;
        (nTop+1)*FONT_SIZE_Y                                                  ,;
        (nRight-1)*FONT_SIZE_X                                                ,;
        (nBottom-1)*FONT_SIZE_Y                                                ;
        SCROLLBAR                                                              ;
        STYLE BUTTON_TYPE_SCROLL_VERT                                          ;
        ACTIONUP  { || NIL }                                                   ;
        ACTIONDN  { || NIL }                                                   ;
        ACTIONBAR { |nSens,nPercent| ImageVertBar(nSens,nPercent,oBrowseImg) } ;
        HANDLE aVertScroll                                                     ;
        ATTACH aWinButtons


                                // -LLIBG- Add horizontal scroll bar
      @ (nLeft+2)*FONT_SIZE_X                                                  ,;
        (nBottom-1)*FONT_SIZE_Y                                                ,;
        (nRight-3)*FONT_SIZE_X                                                 ,;
        (nBottom-0)*FONT_SIZE_Y                                                 ;
        SCROLLBAR                                                               ;
        STYLE BUTTON_TYPE_SCROLL_HORI                                           ;
        ACTIONUP  { || NIL }                                                    ;
        ACTIONDN  { || NIL }                                                    ;
        ACTIONBAR { |nSens,nPercent| ImageHoriBar(nSens,nPercent,oBrowseImg) }  ;
        HANDLE aHoriScroll                                                      ;
        ATTACH aWinButtons

                             // Add a close window button
      @ nLeft*FONT_SIZE_X                       ,;
        nTop *FONT_SIZE_Y                      ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_KEY                    ;
        ACTION K_ESC                             ;
        DISPLAYUP aBmpBase[CLOSWIN_UP]           ;
        DISPLAYDN aBmpBase[CLOSWIN_DW]           ;
        ATTACH aWinButtons


      IF gMode()[LLG_MODE_COLOR_MAX]==256
                             // In 256 colors mode, add a button to manage
                             // exclusive / shared palette
         @ oBrowseImg:nWinDispX2 + 1                               ,;
           oBrowseImg:nWinDispY2 + 1                               ,;
           oBrowseImg:nWinDispX2 + 16                              ,;
           oBrowseImg:nWinDispY2 + 16                               ;
           BUTTON                                                   ;
           STYLE BUTTON_TYPE_RELEASE                                ;
           ACTION { || PalExcShaSwitch(oBrowseImg) }                ;
           DISPLAYUP { || PalExcShaDisp(oBrowseImg,.T.) }           ;
           DISPLAYDN { || PalExcShaDisp(oBrowseImg,.F.) }           ;
           ATTACH aWinButtons

      ENDIF




                                // -LLIBG- Add an icon bar


                                // -LLIBG- Add a button to allow user
                                //         to quit
      @ (nLeft+2 )*FONT_SIZE_X                  ,;
        (nTop+1  )*FONT_SIZE_Y                 ,,;
        BUTTON                                   ;
        STYLE BUTTON_TYPE_KEY                    ;
        ACTION K_ESC                             ;
        ACCELERATOR "Q"                          ;
        DISPLAYUP gBmpLoad("QUIT.BMU")           ;
        DISPLAYDN gBmpLoad("QUIT.BMD")           ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to fit image into
                                //         the window
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgZoom(ASC('/'),oBrowseImg) }                         ;
        ACCELERATOR "S"                                                    ;
        DISPLAYUP gBmpLoad("SMALEST.BMU")                                  ;
        DISPLAYDN gBmpLoad("SMALEST.BMD")                                  ;
        ATTACH aWinButtons

                                // -LLIBG- Add a button to zoom a part of
                                //         the image
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgZoom(ASC(''),oBrowseImg) }                         ;
        ACCELERATOR "A"                                                    ;
        DISPLAYUP gBmpLoad("AREA.BMU")                                     ;
        DISPLAYDN gBmpLoad("AREA.BMD")                                     ;
        ATTACH aWinButtons


                                // -LLIBG- Add a button to rotate to the right
                                //         the image
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgRotate(90,oBrowseImg) }                             ;
        ACCELERATOR "9"                                                    ;
        DISPLAYUP gBmpLoad("ROTAT090.BMU")                                 ;
        DISPLAYDN gBmpLoad("ROTAT090.BMD")                                 ;
        ATTACH aWinButtons

                                // -LLIBG- Add a button to flip the image upside
                                //         down
      @ ATAIL(aWinButton)[BUTTON_RIGHT]+1                                 ,;
        ATAIL(aWinButton)[BUTTON_TOP]                                    ,,;
        BUTTON                                                             ;
        STYLE BUTTON_TYPE_RELEASE                                          ;
        ACTION { || ImgRotate(180,oBrowseImg) }                            ;
        ACCELERATOR "1"                                                    ;
        DISPLAYUP gBmpLoad("ROTAT180.BMU")                                 ;
        DISPLAYDN gBmpLoad("ROTAT180.BMD")                                 ;
        ATTACH aWinButtons


                                // -LLIBG- Show all buttons defined
      SHOW ALL BUTTONS aWinButtons


      // Main loop
      WHILE lMore

         IF !oBrowseImg:lIsStable .AND. NEXTKEY()==0

            ImgDynaDisplay(oBrowseImg)

         ENDIF

         IF oBrowseImg:lIsStable .AND. NEXTKEY()==0
                             //         When the browse move, with K_UP,K_DOWN...
                             //         we need to refresh scrollbars
                             //         locations. Refresh only if no key pending
            ScrBarUpDate(aHoriScroll,ImageHoriBar(NIL,NIL,oBrowseImg))
            ScrBarUpDate(aVertScroll,ImageVertBar(NIL,NIL,oBrowseImg))

         ENDIF

         // Everything's done -- just wait for a key
         nKey := mInkey( 0 , aWinButtons )
                             // -LLIBG- Use mInkey() instead of INKEY() and
                             //         add aWinButtons allow buttons management


         IF ( nKey == K_ESC )
            // Esc means leave
            lMore := .F.

         ELSEIF ( nKey == K_BUTTON )   .OR. ;
                ( nKey == K_CLIC_OUT ) .OR. ;
                ( nKey == K_ACCELERATOR )

            // -LLIBG- Warning, when a button is hitted, mInkey() give the hand
            //         back to main calling routine, and a value of K_BUTTON is
            //         returned. We must not apply this K_BUTTON to applyKey
            //         K_CLIC_OUT is returned when the clic does not match
            //         any button. K_ACCELERATOR is returned when an accelerator
            //         key have been striked
            //         We dont need to apply these keys

         ELSE

            // Apply the key to the ImageBrowse
            EVAL( oBrowseImg:bApplyKey, oBrowseImg, nKey )

         ENDIF

      ENDDO

   ENDIF

                             // Erase VMM item
   oBrowseImg:aDisPtr := iErase(oBrowseImg:aDisPtr)
   oBrowseImg:aOriPtr := iErase(oBrowseImg:aOriPtr)

                             // General restoring
   RESTSCREEN( nTop, nLeft, nBottom, nRight, xScreen)
   SETCOLOR(cColorSave)
   SETCURSOR(nCursoSave)

   RETURN (NIL)




